import numpy as np
import pytest

from autolens.data.imaging import image
from autolens.data.array.util import grid_util
from autolens.data.array.util import mapping_util, mask_util
from autolens.data.array import grids, mask
from autolens.model.inversion import pixelizations as pix


@pytest.fixture(name="msk")
def make_mask():
    return mask.Mask(np.array([[True, False, True],
                               [False, False, False],
                               [True, False, True]]), pixel_scale=1.0)


@pytest.fixture(name="centre_mask")
def make_centre_mask():
    return mask.Mask(np.array([[True, True, True],
                               [True, False, True],
                               [True, True, True]]), pixel_scale=1.0)


@pytest.fixture(name="sub_grid")
def make_sub_grid(msk):
    return grids.SubGrid.from_mask_and_sub_grid_size(msk, sub_grid_size=1)


@pytest.fixture(name="data_grids")
def make_data_grids(centre_mask):
    return grids.DataGrids.grids_from_mask_sub_grid_size_and_psf_shape(centre_mask, 2, (3, 3))


class TestRegularGrid:

    def test__image_grid_from_mask__compare_to_array_util(self):
        msk = np.array([[True, True, False, False],
                        [True, False, True, True],
                        [True, True, False, False]])
        msk = mask.Mask(array=msk, pixel_scale=2.0)

        image_grid_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(mask=msk,
                                                                                             pixel_scales=(2.0, 2.0))

        image_grid = grids.RegularGrid.from_mask(msk)

        assert type(image_grid) == grids.RegularGrid
        assert image_grid == pytest.approx(image_grid_util, 1e-4)
        assert image_grid.pixel_scale == 2.0
        assert (image_grid.mask.masked_grid_index_to_pixel == msk.masked_grid_index_to_pixel).all()

    def test__image_grid_unlensed_grid_properties_compare_to_array_util(self):
        msk = np.array([[True, True, False, False],
                        [True, False, True, True],
                        [True, True, False, False]])
        msk = mask.Mask(array=msk, pixel_scale=2.0)

        image_grid = grids.RegularGrid(arr=np.array([[1.0, 1.0], [1.0, 1.0]]), mask=msk)

        image_grid_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(mask=msk,
                                                                                             pixel_scales=(2.0, 2.0))

        assert type(image_grid) == grids.RegularGrid
        assert image_grid.unlensed_grid == pytest.approx(image_grid_util, 1e-4)

        image_grid_util = grid_util.regular_grid_1d_from_shape_pixel_scales_and_origin(shape=(3, 4), pixel_scales=(2.0, 2.0))

        assert type(image_grid) == grids.RegularGrid
        assert image_grid.unlensed_unmasked_grid == pytest.approx(image_grid_util, 1e-4)

    def test__image_grid_from_shape_and_pixel_scale__compare_to_array_util(self):

        msk = np.array([[False, False, False, False],
                        [False, False, False, False],
                        [False, False, False, False]])
        msk = mask.Mask(array=msk, pixel_scale=2.0)

        image_grid_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(mask=msk,
                                                                                             pixel_scales=(2.0, 2.0))

        image_grid = grids.RegularGrid.from_shape_and_pixel_scale(shape=(3, 4), pixel_scale=2.0)

        assert type(image_grid) == grids.RegularGrid
        assert image_grid == pytest.approx(image_grid_util, 1e-4)
        assert image_grid.pixel_scale == 2.0
        assert (image_grid.mask.masked_grid_index_to_pixel == msk.masked_grid_index_to_pixel).all()

    def test__blurring_grid_from_mask__compare_to_array_util(self):
        msk = np.array([[True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True],
                        [True, True, False, True, True, True, False, True, True],
                        [True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True],
                        [True, True, False, True, True, True, False, True, True],
                        [True, True, True, True, True, True, True, True, True],
                        [True, True, True, True, True, True, True, True, True]])
        msk = mask.Mask(array=msk, pixel_scale=2.0)

        blurring_mask_util = mask_util.mask_blurring_from_mask_and_psf_shape(msk, psf_shape=(3, 5))
        blurring_grid_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(blurring_mask_util,
                                                                                                pixel_scales=(2.0, 2.0))

        msk = mask.Mask(msk, pixel_scale=2.0)
        blurring_grid = grids.RegularGrid.blurring_grid_from_mask_and_psf_shape(mask=msk, psf_shape=(3, 5))

        blurring_msk = msk.blurring_mask_for_psf_shape(psf_shape=(3, 5))

        assert type(blurring_grid) == grids.RegularGrid
        assert blurring_grid == pytest.approx(blurring_grid_util, 1e-4)
        assert blurring_grid.pixel_scale == 2.0
        assert (blurring_grid.mask.masked_grid_index_to_pixel == blurring_msk.masked_grid_index_to_pixel).all()

    def test__map_to_2d__compare_to_util(self):

        msk = np.array([[True, True, False, False],
                        [True, False, True, True],
                        [True, True, False, False]])
        array_1d = np.array([1.0, 6.0, 4.0, 5.0, 2.0])
        one_to_two = np.array([[0, 2], [0, 3], [1, 1], [2, 2], [2, 3]])
        array_2d_util = mapping_util.map_masked_1d_array_to_2d_array_from_array_1d_shape_and_one_to_two(
            array_1d=array_1d, shape=(3, 4), one_to_two=one_to_two)

        msk = mask.Mask(array=msk, pixel_scale=2.0)
        image_grid = grids.RegularGrid.from_mask(msk)
        array_2d_grid = image_grid.scaled_array_from_array_1d(array_1d)

        assert (array_2d_util == array_2d_grid).all()
        assert array_2d_grid.pixel_scale == 2.0
        assert array_2d_grid.origin == (0.0, 0.0)

    def test__scaled_array_from_array_1d__compare_to_util(self):
        msk = np.array([[True, True, False, False],
                        [True, False, True, True],
                        [True, True, False, False]])
        array_1d = np.array([1.0, 6.0, 4.0, 5.0, 2.0])
        one_to_two = np.array([[0, 2], [0, 3], [1, 1], [2, 2], [2, 3]])
        array_2d_util = mapping_util.map_masked_1d_array_to_2d_array_from_array_1d_shape_and_one_to_two(
            array_1d=array_1d, shape=(3, 4), one_to_two=one_to_two)

        msk = mask.Mask(array=msk, pixel_scale=3.0)
        image_grid = grids.RegularGrid.from_mask(msk)
        scaled_array_2d = image_grid.scaled_array_from_array_1d(array_1d)

        assert (scaled_array_2d == array_2d_util).all()
        assert (scaled_array_2d.xticks == np.array([-6.0, -2.0, 2.0, 6.0])).all()
        assert (scaled_array_2d.yticks == np.array([-4.5, -1.5, 1.5, 4.5])).all()
        assert scaled_array_2d.shape_arc_seconds == (9.0, 12.0)
        assert scaled_array_2d.pixel_scale == 3.0
        assert scaled_array_2d.origin == (0.0, 0.0)

    def test__yticks(self):

        sca = grids.RegularGrid(arr=np.array([[1.5, 1.0], [-1.5, -1.0]]), mask=None)
        assert sca.yticks == pytest.approx(np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3)

        sca = grids.RegularGrid(arr=np.array([[3.0, 1.0], [-3.0, -1.0]]), mask=None)
        assert sca.yticks == pytest.approx(np.array([-3.0, -1, 1.0, 3.0]), 1e-3)

        sca = grids.RegularGrid(arr=np.array([[5.0, 3.5], [2.0, -1.0]]), mask=None)
        assert sca.yticks == pytest.approx(np.array([2.0, 3.0, 4.0, 5.0]), 1e-3)

    def test__xticks(self):

        sca = grids.RegularGrid(arr=np.array([[1.0, 1.5], [-1.0, -1.5]]), mask=None)
        assert sca.xticks == pytest.approx(np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3)

        sca = grids.RegularGrid(arr=np.array([[1.0, 3.0], [-1.0, -3.0]]), mask=None)
        assert sca.xticks == pytest.approx(np.array([-3.0, -1, 1.0, 3.0]), 1e-3)

        sca = grids.RegularGrid(arr=np.array([[3.5, 2.0], [-1.0, 5.0]]), mask=None)
        assert sca.xticks == pytest.approx(np.array([2.0, 3.0, 4.0, 5.0]), 1e-3)

    def test__masked_shape_arcsec(self):

        sca = grids.RegularGrid(arr=np.array([[1.5, 1.0], [-1.5, -1.0]]), mask=None)
        assert sca.masked_shape_arcsec == (3.0, 2.0)

        sca = grids.RegularGrid(arr=np.array([[1.5, 1.0], [-1.5, -1.0], [0.1, 0.1]]), mask=None)
        assert sca.masked_shape_arcsec == (3.0, 2.0)

        sca = grids.RegularGrid(arr=np.array([[1.5, 1.0], [-1.5, -1.0], [3.0, 3.0]]), mask=None)
        assert sca.masked_shape_arcsec == (4.5, 4.0)

        sca = grids.RegularGrid(arr=np.array([[1.5, 1.0], [-1.5, -1.0], [3.0, 3.0], [7.0, -5.0]]), mask=None)
        assert sca.masked_shape_arcsec == (8.5, 8.0)


class TestSubGrid(object):

    def test_sub_grid(self, sub_grid):

        assert type(sub_grid) == grids.SubGrid
        assert sub_grid.shape == (5, 2)
        assert sub_grid.pixel_scale == 1.0
        assert (sub_grid == np.array([[1, 0], [0, -1], [0, 0], [0, 1], [-1, 0]])).all()

    def test__image_grid_unlensed_grid_properties_compare_to_array_util(self, msk, sub_grid):

        assert type(sub_grid) == grids.SubGrid
        assert sub_grid.unlensed_grid == \
               pytest.approx(np.array([[1, 0], [0, -1], [0, 0], [0, 1], [-1, 0]]), 1e-4)

        sub_grid_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
            mask=np.full((3, 3), False), pixel_scales=(1.0, 1.0), sub_grid_size=1)

        assert type(sub_grid) == grids.SubGrid
        assert sub_grid.unlensed_unmasked_grid == pytest.approx(sub_grid_util, 1e-4)

        sub_grid = grids.SubGrid.from_mask_and_sub_grid_size(msk, sub_grid_size=2)

        sub_grid_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
            mask=np.full((3, 3), False), pixel_scales=(1.0, 1.0), sub_grid_size=2)

        assert type(sub_grid) == grids.SubGrid
        assert sub_grid.unlensed_unmasked_grid == pytest.approx(sub_grid_util, 1e-4)

    def test_sub_to_pixel(self, sub_grid):
        assert (sub_grid.sub_to_regular == np.array(range(5))).all()

    def test__from_mask(self):
        msk = np.array([[True, True, True],
                        [True, False, False],
                        [True, True, False]])

        sub_grid_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(mask=msk,
                                                                                              pixel_scales=(
                                                                                                  3.0, 3.0),
                                                                                              sub_grid_size=2)

        msk = mask.Mask(msk, pixel_scale=3.0)

        sub_grid = grids.SubGrid.from_mask_and_sub_grid_size(msk, sub_grid_size=2)

        assert sub_grid == pytest.approx(sub_grid_util, 1e-4)

    def test__from_shape_and_pixel_scale(self):
        msk = np.array([[False, False, False],
                        [False, False, False],
                        [False, False, False]])

        sub_grid_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(mask=msk,
                                                                                              pixel_scales=(
                                                                                                  3.0, 3.0),
                                                                                              sub_grid_size=2)

        sub_grid = grids.SubGrid.from_shape_pixel_scale_and_sub_grid_size(shape=(3, 3), pixel_scale=3.0, sub_grid_size=2)

        assert sub_grid == pytest.approx(sub_grid_util, 1e-4)

    def test__map_to_2d__compare_to_util(self):
        msk = np.array([[True, True, False, False],
                        [True, False, True, True],
                        [True, True, False, False]])
        array_1d = np.array([1.0, 6.0, 4.0, 5.0, 2.0])
        one_to_two = np.array([[0, 2], [0, 3], [1, 1], [2, 2], [2, 3]])
        array_2d_util = mapping_util.map_masked_1d_array_to_2d_array_from_array_1d_shape_and_one_to_two(
            array_1d=array_1d,
            shape=(3, 4), one_to_two=one_to_two)

        msk = mask.Mask(array=msk, pixel_scale=2.0)
        image_grid = grids.SubGrid.from_mask_and_sub_grid_size(msk, sub_grid_size=2)
        array_2d_grid = image_grid.scaled_array_from_array_1d(array_1d)

        assert (array_2d_util == array_2d_grid).all()
        assert array_2d_grid.pixel_scale == 2.0
        assert array_2d_grid.origin == (0.0, 0.0)

    def test_sub_data_to_image(self, sub_grid):

        assert (sub_grid.sub_data_to_regular_data(np.array(range(5))) == np.array(range(5))).all()

    def test_sub_to_image__compare_to_array_util(self):

        msk = np.array([[True, False, True],
                        [False, False, False],
                        [True, False, False]])

        sub_to_image_util = mapping_util.sub_to_regular_from_mask(msk, sub_grid_size=2)

        msk = mask.Mask(msk, pixel_scale=3.0)

        sub_grid = grids.SubGrid.from_mask_and_sub_grid_size(mask=msk, sub_grid_size=2)
        assert sub_grid.sub_grid_size == 2
        assert sub_grid.sub_grid_fraction == (1.0 / 4.0)
        assert (sub_grid.sub_to_regular == sub_to_image_util).all()


class TestPixGrid:

    def test_pix_image_grid__attributes(self):

        pix_image_grid = grids.PixGrid(arr=np.array([[1.0, 1.0], [2.0, 2.0]]),
                                       regular_to_nearest_regular_pix=np.array([0, 1]))

        assert type(pix_image_grid) == grids.PixGrid
        assert (pix_image_grid == np.array([[1.0, 1.0], [2.0, 2.0]])).all()
        assert (pix_image_grid.regular_to_nearest_regular_pix == np.array([0, 1])).all()


class TestSparseToRegularGrid:

    def test__properties_consistent_with_mapping_util(self):

        ma = mask.Mask(array=np.array([[True, False, True],
                                       [False, False, False],
                                       [True, False, True]]), pixel_scale=0.5)

        image_grid = grids.RegularGrid.from_mask(mask=ma)

        pix_grid = grids.SparseToRegularGrid(unmasked_sparse_grid_shape=(10, 10), pixel_scales=(0.16, 0.16),
                                             regular_grid=image_grid)

        full_pix_grid_pixel_centres = image_grid.mask.grid_arc_seconds_to_grid_pixel_centres(pix_grid.unmasked_sparse_grid)
        total_pix_pixels = mask_util.total_sparse_pixels_from_mask(mask=ma,
                                                                   unmasked_sparse_grid_pixel_centres=full_pix_grid_pixel_centres)

        image_to_full_pix_util = pix_grid.grid_arc_seconds_to_grid_pixel_indexes(grid_arc_seconds=image_grid)

        pix_to_full_pix_util = mapping_util.sparse_to_unmasked_sparse_from_mask_and_pixel_centres(
            total_sparse_pixels=total_pix_pixels, mask=ma, unmasked_sparse_grid_pixel_centres=full_pix_grid_pixel_centres).astype('int')

        full_pix_to_pix_util = mapping_util.unmasked_sparse_to_sparse_from_mask_and_pixel_centres(mask=ma,
                                        unmasked_sparse_grid_pixel_centres=full_pix_grid_pixel_centres,
                                         total_sparse_pixels=pix_grid.total_sparse_pixels).astype('int')

        image_to_pix_util = mapping_util.regular_to_sparse_from_sparse_mappings(
            regular_to_unmasked_sparse=image_to_full_pix_util,
            unmasked_sparse_to_sparse=full_pix_to_pix_util)

        pix_grid_util = mapping_util.sparse_grid_from_unmasked_sparse_grid(unmasked_sparse_grid=pix_grid.grid_1d,
                                                                           sparse_to_unmasked_sparse=pix_to_full_pix_util)

        assert pix_grid.total_sparse_pixels == total_pix_pixels
        assert (pix_grid.sparse_to_unmasked_sparse == pix_to_full_pix_util).all()
        assert (pix_grid.unmasked_sparse_to_sparse == full_pix_to_pix_util).all()
        assert (pix_grid.regular_to_unmasked_sparse == image_to_full_pix_util).all()
        assert (pix_grid.regular_to_sparse == image_to_pix_util).all()
        assert (pix_grid.sparse_grid == pix_grid_util).all()

    def test__pixelization_grid_overlaps_mask_perfectly__masked_pixels_in_masked_pixelization_grid(self):

            ma = mask.Mask(array=np.array([[True, False, True],
                                           [False, False, False],
                                           [True, False, True]]), pixel_scale=1.0)

            image_grid = grids.RegularGrid.from_mask(mask=ma)

            pix_grid = grids.SparseToRegularGrid(unmasked_sparse_grid_shape=(3, 3), pixel_scales=(1.0, 1.0),
                                                 regular_grid=image_grid)

            assert pix_grid.total_sparse_pixels == 5
            assert (pix_grid.sparse_to_unmasked_sparse == np.array([1, 3, 4, 5, 7])).all()
            assert (pix_grid.unmasked_sparse_to_sparse == np.array([0, 0, 1, 1, 2, 3, 4, 4, 4])).all()
            assert (pix_grid.regular_to_unmasked_sparse == np.array([1, 3, 4, 5, 7])).all()
            assert (pix_grid.regular_to_sparse == np.array([0, 1, 2, 3, 4])).all()
            assert (pix_grid.sparse_grid == np.array([[1.0, 0.0], [0.0, -1.0], [0.0, 0.0], [0.0, 1.0],
                                                      [-1.0, 0.0]])).all()

    def test__same_as_above_but_4x3_grid_and_mask(self):

            ma = mask.Mask(array=np.array([[True, False, True],
                                           [False, False, False],
                                           [False, False, False],
                                           [True, False, True]]), pixel_scale=1.0)

            image_grid = grids.RegularGrid.from_mask(mask=ma)

            pix_grid = grids.SparseToRegularGrid(unmasked_sparse_grid_shape=(4, 3), pixel_scales=(1.0, 1.0),
                                                 regular_grid=image_grid)

            assert pix_grid.total_sparse_pixels == 8
            assert (pix_grid.sparse_to_unmasked_sparse == np.array([1, 3, 4, 5, 6, 7, 8, 10])).all()
            assert (pix_grid.unmasked_sparse_to_sparse == np.array([0, 0, 1, 1, 2, 3, 4, 5, 6, 7, 7, 7])).all()
            assert (pix_grid.regular_to_unmasked_sparse == np.array([1, 3, 4, 5, 6, 7, 8, 10])).all()
            assert (pix_grid.regular_to_sparse == np.array([0, 1, 2, 3, 4, 5, 6, 7])).all()
            assert (pix_grid.sparse_grid == np.array([[1.5, 0.0],
                                                      [ 0.5, -1.0], [ 0.5, 0.0], [ 0.5, 1.0],
                                                      [-0.5, -1.0], [-0.5, 0.0], [-0.5, 1.0],
                                                      [-1.5, 0.0]])).all()

    def test__same_as_above_but_3x4_grid_and_mask(self):

            ma = mask.Mask(array=np.array([[True, False, True, True],
                                           [False, False, False, False],
                                           [True,  False,  True,  True]]), pixel_scale=1.0)

            image_grid = grids.RegularGrid.from_mask(mask=ma)

            pix_grid = grids.SparseToRegularGrid(unmasked_sparse_grid_shape=(3, 4), pixel_scales=(1.0, 1.0),
                                                 regular_grid=image_grid)

            assert pix_grid.total_sparse_pixels == 6
            assert (pix_grid.sparse_to_unmasked_sparse == np.array([1, 4, 5, 6, 7, 9])).all()
            assert (pix_grid.unmasked_sparse_to_sparse == np.array([0, 0, 1, 1, 1, 2, 3, 4, 5, 5, 5, 5])).all()
            assert (pix_grid.regular_to_unmasked_sparse == np.array([1, 4, 5, 6, 7, 9])).all()
            assert (pix_grid.regular_to_sparse == np.array([0, 1, 2, 3, 4, 5])).all()
            assert (pix_grid.sparse_grid == np.array([[1.0, -0.5],
                                                      [0.0, -1.5], [0.0, -0.5], [0.0, 0.5], [0.0, 1.5],
                                                      [-1.0, -0.5]])).all()


class TestUnmaskedGrids:
    class TestPaddedImageGridFromShapes:

        def test__3x3_array__psf_size_is_1x1__no_padding(self):
            image_padded_grid = grids.PaddedRegularGrid.padded_grid_from_shapes_and_pixel_scale(shape=(3, 3),
                                                                                                psf_shape=(1, 1),
                                                                                                pixel_scale=1.0)

            assert len(image_padded_grid) == 9
            assert image_padded_grid.pixel_scale == 1.0
            assert image_padded_grid.image_shape == (3, 3)
            assert image_padded_grid.padded_shape == (3, 3)

        def test__3x3_image__5x5_psf_size__7x7_image_padded_grid_made(self):
            image_padded_grid = grids.PaddedRegularGrid.padded_grid_from_shapes_and_pixel_scale(shape=(3, 3),
                                                                                                psf_shape=(5, 5),
                                                                                                pixel_scale=1.0)

            assert len(image_padded_grid) == 49
            assert image_padded_grid.image_shape == (3, 3)
            assert image_padded_grid.padded_shape == (7, 7)

        def test__3x3_image__7x7_psf_size__9x9_image_padded_grid_made(self):
            image_padded_grid = grids.PaddedRegularGrid.padded_grid_from_shapes_and_pixel_scale(shape=(3, 3),
                                                                                                psf_shape=(7, 7),
                                                                                                pixel_scale=1.0)
            assert len(image_padded_grid) == 81
            assert image_padded_grid.image_shape == (3, 3)
            assert image_padded_grid.padded_shape == (9, 9)

        def test__4x3_image__3x3_psf_size__6x5_image_padded_grid_made(self):
            image_padded_grid = grids.PaddedRegularGrid.padded_grid_from_shapes_and_pixel_scale(shape=(4, 3),
                                                                                                psf_shape=(3, 3),
                                                                                                pixel_scale=1.0)
            assert len(image_padded_grid) == 30
            assert image_padded_grid.image_shape == (4, 3)
            assert image_padded_grid.padded_shape == (6, 5)

        def test__3x4_image__3x3_psf_size__5x6_image_padded_grid_made(self):
            image_padded_grid = grids.PaddedRegularGrid.padded_grid_from_shapes_and_pixel_scale(shape=(3, 4),
                                                                                                psf_shape=(3, 3),
                                                                                                pixel_scale=1.0)

            assert len(image_padded_grid) == 30
            assert image_padded_grid.image_shape == (3, 4)
            assert image_padded_grid.padded_shape == (5, 6)

        def test__4x4_image__3x3_psf_size__6x6_image_padded_grid_made(self):
            image_padded_grid = grids.PaddedRegularGrid.padded_grid_from_shapes_and_pixel_scale(shape=(4, 4),
                                                                                                psf_shape=(3, 3),
                                                                                                pixel_scale=1.0)

            assert len(image_padded_grid) == 36
            assert image_padded_grid.image_shape == (4, 4)
            assert image_padded_grid.padded_shape == (6, 6)

        def test__image_padded_grid_coordinates__match_grid_2d_after_padding(self):
            image_padded_grid = grids.PaddedRegularGrid.padded_grid_from_shapes_and_pixel_scale(shape=(4, 4),
                                                                                                psf_shape=(3, 3),
                                                                                                pixel_scale=3.0)

            image_padded_grid_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(
                mask=np.full((6, 6), False), pixel_scales=(3.0, 3.0))
            assert (image_padded_grid == image_padded_grid_util).all()
            assert image_padded_grid.image_shape == (4, 4)
            assert image_padded_grid.padded_shape == (6, 6)

            image_padded_grid = grids.PaddedRegularGrid.padded_grid_from_shapes_and_pixel_scale(shape=(4, 5),
                                                                                                psf_shape=(3, 3),
                                                                                                pixel_scale=2.0)
            image_padded_grid_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(
                mask=np.full((6, 7), False), pixel_scales=(2.0, 2.0))
            assert (image_padded_grid == image_padded_grid_util).all()
            assert image_padded_grid.image_shape == (4, 5)
            assert image_padded_grid.padded_shape == (6, 7)

            image_padded_grid = grids.PaddedRegularGrid.padded_grid_from_shapes_and_pixel_scale(shape=(5, 4),
                                                                                                psf_shape=(3, 3),
                                                                                                pixel_scale=1.0)
            image_padded_grid_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(
                mask=np.full((7, 6), False), pixel_scales=(1.0, 1.0))
            assert (image_padded_grid == image_padded_grid_util).all()
            assert image_padded_grid.image_shape == (5, 4)
            assert image_padded_grid.padded_shape == (7, 6)

            image_padded_grid = grids.PaddedRegularGrid.padded_grid_from_shapes_and_pixel_scale(shape=(2, 5),
                                                                                                psf_shape=(5, 5),
                                                                                                pixel_scale=8.0)
            image_padded_grid_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(
                mask=np.full((6, 9), False), pixel_scales=(8.0, 8.0))
            assert (image_padded_grid == image_padded_grid_util).all()
            assert image_padded_grid.image_shape == (2, 5)
            assert image_padded_grid.padded_shape == (6, 9)

    class TestPaddedSubGridFromMask:

        def test__3x3_array__psf_size_is_1x1__no_padding(self):
            msk = mask.Mask(array=np.full((3, 3), False), pixel_scale=1.0)

            sub_padded_grid = grids.PaddedSubGrid.padded_grid_from_mask_sub_grid_size_and_psf_shape(mask=msk,
                                                                                                    sub_grid_size=3,
                                                                                                    psf_shape=(1, 1))

            assert len(sub_padded_grid) == 9 * 3 ** 2
            assert sub_padded_grid.image_shape == (3, 3)
            assert sub_padded_grid.padded_shape == (3, 3)

        def test__3x3_image__5x5_psf_size__7x7_image_grid_made(self):
            msk = mask.Mask(array=np.full((3, 3), False), pixel_scale=1.0)

            sub_padded_grid = grids.PaddedSubGrid.padded_grid_from_mask_sub_grid_size_and_psf_shape(mask=msk,
                                                                                                    sub_grid_size=2,
                                                                                                    psf_shape=(5, 5))

            assert len(sub_padded_grid) == 49 * 2 ** 2
            assert sub_padded_grid.image_shape == (3, 3)
            assert sub_padded_grid.padded_shape == (7, 7)

        def test__4x3_image__3x3_psf_size__6x5_image_grid_made(self):
            msk = mask.Mask(array=np.full((4, 3), False), pixel_scale=1.0)

            sub_padded_grid = grids.PaddedSubGrid.padded_grid_from_mask_sub_grid_size_and_psf_shape(mask=msk,
                                                                                                    sub_grid_size=2,
                                                                                                    psf_shape=(3, 3))

            assert len(sub_padded_grid) == 30 * 2 ** 2
            assert sub_padded_grid.image_shape == (4, 3)
            assert sub_padded_grid.padded_shape == (6, 5)

        def test__3x4_image__3x3_psf_size__5x6_image_grid_made(self):
            msk = mask.Mask(array=np.full((3, 4), False), pixel_scale=1.0)

            sub_padded_grid = grids.PaddedSubGrid.padded_grid_from_mask_sub_grid_size_and_psf_shape(mask=msk,
                                                                                                    sub_grid_size=2,
                                                                                                    psf_shape=(3, 3))

            assert len(sub_padded_grid) == 30 * 2 ** 2
            assert sub_padded_grid.image_shape == (3, 4)
            assert sub_padded_grid.padded_shape == (5, 6)

        def test__4x4_image__3x3_psf_size__6x6_image_grid_made(self):
            msk = mask.Mask(array=np.full((4, 4), False), pixel_scale=1.0)

            sub_padded_grid = grids.PaddedSubGrid.padded_grid_from_mask_sub_grid_size_and_psf_shape(mask=msk,
                                                                                                    sub_grid_size=4,
                                                                                                    psf_shape=(3, 3))

            assert len(sub_padded_grid) == 36 * 4 ** 2
            assert sub_padded_grid.image_shape == (4, 4)
            assert sub_padded_grid.padded_shape == (6, 6)

        def test__sub_padded_grid_coordinates__match_grid_2d_after_padding(self):
            msk = mask.Mask(array=np.full((4, 4), False), pixel_scale=3.0)

            sub_padded_grid = grids.PaddedSubGrid.padded_grid_from_mask_sub_grid_size_and_psf_shape(mask=msk,
                                                                                                    sub_grid_size=3,
                                                                                                    psf_shape=(3, 3))

            sub_padded_grid_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
                mask=np.full((6, 6), False), pixel_scales=(3.0, 3.0), sub_grid_size=3)

            assert sub_padded_grid == pytest.approx(sub_padded_grid_util, 1e-4)

            msk = mask.Mask(array=np.full((4, 5), False), pixel_scale=2.0)

            sub_padded_grid = grids.PaddedSubGrid.padded_grid_from_mask_sub_grid_size_and_psf_shape(mask=msk,
                                                                                                    sub_grid_size=1,
                                                                                                    psf_shape=(3, 3))

            sub_padded_grid_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
                mask=np.full((6, 7), False), pixel_scales=(2.0, 2.0), sub_grid_size=1)

            assert sub_padded_grid == pytest.approx(sub_padded_grid_util, 1e-4)

            msk = mask.Mask(array=np.full((5, 4), False), pixel_scale=2.0)

            sub_padded_grid = grids.PaddedSubGrid.padded_grid_from_mask_sub_grid_size_and_psf_shape(mask=msk,
                                                                                                    sub_grid_size=2,
                                                                                                    psf_shape=(3, 3))

            sub_padded_grid_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
                mask=np.full((7, 6), False), pixel_scales=(2.0, 2.0), sub_grid_size=2)

            assert sub_padded_grid == pytest.approx(sub_padded_grid_util, 1e-4)

            msk = mask.Mask(array=np.full((2, 5), False), pixel_scale=8.0)

            sub_padded_grid = grids.PaddedSubGrid.padded_grid_from_mask_sub_grid_size_and_psf_shape(mask=msk,
                                                                                                    sub_grid_size=4,
                                                                                                    psf_shape=(5, 5))

            sub_padded_grid_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
                mask=np.full((6, 9), False), pixel_scales=(8.0, 8.0), sub_grid_size=4)

            assert sub_padded_grid == pytest.approx(sub_padded_grid_util, 1e-4)

    class TestConvolve:

        def test__convolve_1d_mapper_array_with_psf_and_trim_to_original_size(self):
            msk = mask.Mask(array=np.full((4, 4), False), pixel_scale=1.0)

            image_padded_grid = grids.PaddedRegularGrid(arr=np.empty((0)), mask=msk, image_shape=(2, 2))

            array_1d = np.array([0.0, 0.0, 0.0, 0.0,
                                 0.0, 0.0, 0.0, 0.0,
                                 0.0, 1.0, 0.0, 0.0,
                                 0.0, 0.0, 0.0, 0.0])

            kernel = np.array([[0.0, 1.0, 0.0],
                               [1.0, 2.0, 1.0],
                               [0.0, 1.0, 0.0]])

            psf = image.PSF(array=kernel, pixel_scale=1.0)

            blurred_array_1d = image_padded_grid.convolve_array_1d_with_psf(array_1d, psf)

            assert (blurred_array_1d == np.array([0.0, 0.0, 0.0, 0.0,
                                                  0.0, 1.0, 0.0, 0.0,
                                                  1.0, 2.0, 1.0, 0.0,
                                                  0.0, 1.0, 0.0, 0.0])).all()

        def test__same_as_above_but_different_quantities(self):
            msk = mask.Mask(array=np.full((5, 4), False), pixel_scale=1.0)

            image_padded_grid = grids.PaddedRegularGrid(arr=np.empty((0)), mask=msk, image_shape=(3, 2))

            array_1d = np.array([0.0, 0.0, 0.0, 0.0,
                                 0.0, 0.0, 0.0, 0.0,
                                 0.0, 1.0, 0.0, 0.0,
                                 0.0, 0.0, 0.0, 0.0,
                                 1.0, 0.0, 0.0, 0.0])

            kernel = np.array([[1.0, 1.0, 4.0],
                               [1.0, 3.0, 1.0],
                               [1.0, 1.0, 1.0]])

            psf = image.PSF(array=kernel, pixel_scale=1.0)

            blurred_array_1d = image_padded_grid.convolve_array_1d_with_psf(array_1d, psf)

            assert (blurred_array_1d == np.array([0.0, 0.0, 0.0, 0.0,
                                                  1.0, 1.0, 4.0, 0.0,
                                                  1.0, 3.0, 1.0, 0.0,
                                                  2.0, 5.0, 1.0, 0.0,
                                                  3.0, 1.0, 0.0, 0.0])).all()

    class TestMapUnmaskedArrays:

        def test__map_to_2d_keep_padded__4x4_from_1d(self):
            msk = mask.Mask(array=np.full((4, 4), False), pixel_scale=1.0)

            image_padded_grid = grids.PaddedRegularGrid(arr=np.empty((0)), mask=msk, image_shape=(2, 2))

            array_1d = np.array([6.0, 7.0, 9.0, 10.0,
                                 1.0, 2.0, 3.0, 4.0,
                                 5.0, 6.0, 7.0, 8.0,
                                 1.0, 2.0, 3.0, 4.0])
            array_2d = image_padded_grid.map_to_2d_keep_padded(array_1d)

            assert (array_2d == np.array([[6.0, 7.0, 9.0, 10.0],
                                          [1.0, 2.0, 3.0, 4.0],
                                          [5.0, 6.0, 7.0, 8.0],
                                          [1.0, 2.0, 3.0, 4.0]])).all()

        def test__map_to_2d_keep_padded__5x3_from_1d(self):
            msk = mask.Mask(array=np.full((5, 3), False), pixel_scale=1.0)

            image_padded_grid = grids.PaddedRegularGrid(arr=np.empty((0)), mask=msk, image_shape=(3, 1))

            array_1d = np.array([1.0, 2.0, 3.0,
                                 4.0, 5.0, 6.0,
                                 7.0, 8.0, 9.0,
                                 1.0, 2.0, 3.0,
                                 4.0, 5.0, 6.0])
            array_2d = image_padded_grid.map_to_2d_keep_padded(array_1d)

            assert (array_2d == np.array([[1.0, 2.0, 3.0],
                                          [4.0, 5.0, 6.0],
                                          [7.0, 8.0, 9.0],
                                          [1.0, 2.0, 3.0],
                                          [4.0, 5.0, 6.0]])).all()

        def test__map_to_2d_keep_padded__3x5__from_1d(self):
            msk = mask.Mask(array=np.full((3, 5), False), pixel_scale=1.0)

            image_padded_grid = grids.PaddedRegularGrid(arr=np.empty((0)), mask=msk, image_shape=(1, 3))

            array_1d = np.array([1.0, 2.0, 3.0, 4.0, 5.0,
                                 6.0, 7.0, 8.0, 9.0, 1.0,
                                 2.0, 3.0, 4.0, 5.0, 6.0])
            array_2d = image_padded_grid.map_to_2d_keep_padded(array_1d)

            assert (array_2d == np.array([[1.0, 2.0, 3.0, 4.0, 5.0],
                                          [6.0, 7.0, 8.0, 9.0, 1.0],
                                          [2.0, 3.0, 4.0, 5.0, 6.0]])).all()

        def test__map_to_2d_and_trim__4x4_to_2x2__from_1d(self):
            msk = mask.Mask(array=np.full((4, 4), False), pixel_scale=1.0)

            image_padded_grid = grids.PaddedRegularGrid(arr=np.empty((0)), mask=msk, image_shape=(2, 2))

            array_1d = np.array([1.0, 2.0, 3.0, 4.0,
                                 5.0, 6.0, 7.0, 8.0,
                                 9.0, 1.0, 2.0, 3.0,
                                 4.0, 5.0, 6.0, 7.0])
            array_2d = image_padded_grid.map_to_2d(padded_array_1d=array_1d)

            assert (array_2d == np.array([[6.0, 7.0],
                                          [1.0, 2.0]])).all()

        def test__map_to_2d_and_trim__5x3_to_3x1__from_1d(self):
            msk = mask.Mask(array=np.full((5, 3), False), pixel_scale=1.0)

            image_padded_grid = grids.PaddedRegularGrid(arr=np.empty((0)), mask=msk, image_shape=(3, 1))

            array_1d = np.array([1.0, 2.0, 3.0,
                                 4.0, 5.0, 6.0,
                                 7.0, 8.0, 9.0,
                                 1.0, 2.0, 3.0,
                                 4.0, 5.0, 6.0])
            array_2d = image_padded_grid.map_to_2d(array_1d)

            assert (array_2d == np.array([[5.0],
                                          [8.0],
                                          [2.0]])).all()

        def test__map_to_2d_and_trim__3x5_to_1x3__from_1d(self):
            msk = mask.Mask(array=np.full((3, 5), False), pixel_scale=1.0)

            image_padded_grid = grids.PaddedRegularGrid(arr=np.empty((0)), mask=msk, image_shape=(1, 3))

            array_1d = np.array([1.0, 2.0, 3.0, 4.0, 5.0,
                                 6.0, 7.0, 8.0, 9.0, 1.0,
                                 2.0, 3.0, 4.0, 5.0, 6.0])
            array_2d = image_padded_grid.map_to_2d(array_1d)

            assert (array_2d == np.array([[7.0, 8.0, 9.0]])).all()


class TestDataGrids(object):

    def test__grids(self, data_grids):
        assert (data_grids.regular == np.array([[0., 0.]])).all()
        np.testing.assert_almost_equal(data_grids.sub, np.array([[0.16666667, -0.16666667],
                                                                    [0.16666667, 0.16666667],
                                                                    [-0.16666667, -0.16666667],
                                                                    [-0.16666667, 0.16666667]]))
        assert (data_grids.blurring == np.array([[1., -1.],
                                                    [1., 0.],
                                                    [1., 1.],
                                                    [0., -1.],
                                                    [0., 1.],
                                                    [-1., -1.],
                                                    [-1., 0.],
                                                    [-1., 1.]])).all()
        assert (data_grids.pix == np.array([[0.0, 0.0]])).all()

    def test__from_shape_and_pixel_scale(self):

        ma = mask.Mask(np.array([[False, False, False],
                                 [False, False, False],
                                 [False, False, False]]), pixel_scale=2.0)

        data_grids_mask = grids.DataGrids.grids_from_mask_sub_grid_size_and_psf_shape(mask=ma, sub_grid_size=2,
                                                                                         psf_shape=(1, 1))

        data_grids_shape = grids.DataGrids.from_shape_and_pixel_scale(shape=(3, 3), pixel_scale=2.0,
                                                                         sub_grid_size=2)

        assert (data_grids_mask.regular == data_grids_shape.regular).all()
        assert (data_grids_mask.sub == data_grids_shape.sub).all()
        assert (data_grids_mask.pix == np.array([[0.0, 0.0]])).all()

    def test__padded_grids(self):
        msk = np.array([[False, False],
                        [False, False]])

        msk = mask.Mask(msk, pixel_scale=1.0)

        padded_grids = grids.DataGrids.padded_grids_from_mask_sub_grid_size_and_psf_shape(msk, sub_grid_size=2,
                                                                                          psf_shape=(3, 3))

        image_padded_grid_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(
            mask=np.full((4, 4), False),
            pixel_scales=(1.0, 1.0))

        sub_padded_grid_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
            mask=np.full((4, 4), False),
            pixel_scales=(1.0, 1.0), sub_grid_size=2)

        assert (padded_grids.regular == image_padded_grid_util).all()
        assert padded_grids.regular.image_shape == (2, 2)
        assert padded_grids.regular.padded_shape == (4, 4)

        assert padded_grids.sub == pytest.approx(sub_padded_grid_util, 1e-4)
        assert padded_grids.sub.image_shape == (2, 2)
        assert padded_grids.sub.padded_shape == (4, 4)

        assert (padded_grids.blurring == np.array([0.0, 0.0])).all()

        assert (padded_grids.pix == np.array([[0.0, 0.0]])).all()

    def test__for_simulation(self):

        padded_grids = grids.DataGrids.grids_for_simulation(shape=(2, 2), pixel_scale=1.0, sub_grid_size=2,
                                                            psf_shape=(3, 3))

        image_padded_grid_util = grid_util.regular_grid_1d_masked_from_mask_pixel_scales_and_origin(
            mask=np.full((4, 4), False),
            pixel_scales=(1.0, 1.0))

        sub_padded_grid_util = grid_util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(
            mask=np.full((4, 4), False),
            pixel_scales=(1.0, 1.0), sub_grid_size=2)

        assert (padded_grids.regular == image_padded_grid_util).all()
        assert padded_grids.regular.image_shape == (2, 2)
        assert padded_grids.regular.padded_shape == (4, 4)

        assert padded_grids.sub == pytest.approx(sub_padded_grid_util, 1e-4)
        assert padded_grids.sub.image_shape == (2, 2)
        assert padded_grids.sub.padded_shape == (4, 4)

        assert (padded_grids.blurring == np.array([0.0, 0.0])).all()

        assert (padded_grids.pix == np.array([[0.0, 0.0]])).all()

    def test__apply_function_retains_attributes(self, data_grids):

        data_grids.pix = data_grids.regular

        def add_one(coords):
            return np.add(1, coords)

        assert isinstance(data_grids.regular, grids.RegularGrid)
        assert data_grids.regular.mask is not None

        assert isinstance(data_grids.sub, grids.SubGrid)
        assert data_grids.sub.mask is not None
        assert data_grids.sub.sub_grid_size is not None
        assert data_grids.sub.sub_grid_length is not None
        assert data_grids.sub.sub_grid_fraction is not None

        new_collection = data_grids.apply_function(add_one)

        assert new_collection.regular.mask is not None
        assert new_collection.sub.mask is not None
        assert new_collection.sub.sub_grid_size is not None
        assert new_collection.sub.sub_grid_length is not None
        assert new_collection.sub.sub_grid_fraction is not None

        assert isinstance(data_grids.pix, grids.RegularGrid)
        assert data_grids.pix.mask is not None

    def test__apply_function(self, data_grids):

        data_grids.pix = data_grids.regular

        def add_one(coords):
            return np.add(1, coords)

        new_collection = data_grids.apply_function(add_one)
        assert isinstance(new_collection, grids.DataGrids)
        assert (new_collection.regular == np.add(1, np.array([[0., 0.]]))).all()
        np.testing.assert_almost_equal(new_collection.sub, np.add(1, np.array([[0.16666667, -0.16666667],
                                                                               [0.16666667, 0.16666667],
                                                                               [-0.16666667, -0.16666667],
                                                                               [-0.16666667, 0.16666667]])))
        assert (new_collection.blurring == np.add(1, np.array([[1., -1.],
                                                               [1., 0.],
                                                               [1., 1.],
                                                               [0., -1.],
                                                               [0., 1.],
                                                               [-1., -1.],
                                                               [-1., 0.],
                                                               [-1., 1.]]))).all()
        assert (new_collection.pix == np.add(1, np.array([[0., 0.]]))).all()

    def test__map_function(self, data_grids):

        data_grids.pix = data_grids.regular

        def add_number(coords, number):
            return np.add(coords, number)

        new_collection = data_grids.map_function(add_number, [1, 2, 3, 1])

        assert isinstance(new_collection, grids.DataGrids)
        assert (new_collection.regular == np.add(1, np.array([[0., 0.]]))).all()
        np.testing.assert_almost_equal(new_collection.sub, np.add(2, np.array([[0.16666667, -0.16666667],
                                                                               [0.16666667, 0.16666667],
                                                                               [-0.16666667, -0.16666667],
                                                                               [-0.16666667, 0.16666667]])))
        assert (new_collection.blurring == np.add(3, np.array([[1., -1.],
                                                               [1., 0.],
                                                               [1., 1.],
                                                               [0., -1.],
                                                               [0., 1.],
                                                               [-1., -1.],
                                                               [-1., 0.],
                                                               [-1., 1.]]))).all()

        print(new_collection.pix)
        assert (new_collection.pix == np.add(1, np.array([[0., 0.]]))).all()

    def test__data_grids_with_pix_grid(self, data_grids):

        data_grids = data_grids.data_grids_with_pix_grid(pix_grid=np.array([[5.0, 5.0], [6.0, 7.0]]),
                                                               regular_to_nearest_regular_pix=np.array([0, 1]))

        assert (data_grids.regular == np.array([[0., 0.]])).all()
        np.testing.assert_almost_equal(data_grids.sub, np.array([[0.16666667, -0.16666667],
                                                                    [0.16666667, 0.16666667],
                                                                    [-0.16666667, -0.16666667],
                                                                    [-0.16666667, 0.16666667]]))
        assert (data_grids.blurring == np.array([[1., -1.],
                                                    [1., 0.],
                                                    [1., 1.],
                                                    [0., -1.],
                                                    [0., 1.],
                                                    [-1., -1.],
                                                    [-1., 0.],
                                                    [-1., 1.]])).all()
        assert (data_grids.pix == np.array([[5.0, 5.0], [6.0, 7.0]])).all()
        assert (data_grids.pix.regular_to_nearest_regular_pix == np.array([0, 1])).all()


class TestImageGridBorder(object):

    class TestFromMask:

        def test__simple_mask_border_pixels_is_border(self):

            msk = np.array([[False, False, False, False, False, False, False, True],
                            [False, True, True, True, True, True, False, True],
                            [False, True, False, False, False, True, False, True],
                            [False, True, False, True, False, True, False, True],
                            [False, True, False, False, False, True, False, True],
                            [False, True, True, True, True, True, False, True],
                            [False, False, False, False, False, False, False, True]])

            msk = mask.Mask(msk, pixel_scale=3.0)

            border = grids.RegularGridBorder.from_mask(msk)

            assert (border == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 13, 14, 17, 18, 22, 23, 24, 25,
                                        26, 27, 28, 29, 30, 31])).all()

    class TestRelocateCoordinates(object):

        def test__inside_border_no_relocations(self):

            thetas = np.linspace(0.0, 2.0 * np.pi, 32)
            grid_circle = list(map(lambda x: (np.cos(x), np.sin(x)), thetas))
            grid = grid_circle
            grid.append(np.array([0.1, 0.0]))
            grid.append(np.array([-0.2, -0.3]))
            grid.append(np.array([0.5, 0.4]))
            grid.append(np.array([0.7, -0.1]))
            image_grid = np.asarray(grid)
            sub_grid = np.asarray(grid)
            sub_grid[35 ,0] = 0.5
            sub_grid[35 ,1] = 0.3
            data_grids = grids.DataGrids(regular=image_grid, sub=sub_grid, blurring=None)

            border = grids.RegularGridBorder(arr=np.arange(32))
            relocated_grids = border.relocated_grids_from_grids(data_grids)

            assert relocated_grids.regular[0:32] == pytest.approx(np.asarray(grid_circle)[0:32], 1e-3)
            assert relocated_grids.regular[32] == pytest.approx(np.array([0.1, 0.0]), 1e-3)
            assert relocated_grids.regular[33] == pytest.approx(np.array([-0.2, -0.3]), 1e-3)
            assert relocated_grids.regular[34] == pytest.approx(np.array([0.5, 0.4]), 1e-3)
            assert relocated_grids.regular[35] == pytest.approx(np.array([0.7, -0.1]), 1e-3)

            assert relocated_grids.sub[0:32] == pytest.approx(np.asarray(grid_circle)[0:32], 1e-3)
            assert relocated_grids.sub[32] == pytest.approx(np.array([0.1, 0.0]), 1e-3)
            assert relocated_grids.sub[33] == pytest.approx(np.array([-0.2, -0.3]), 1e-3)
            assert relocated_grids.sub[34] == pytest.approx(np.array([0.5, 0.4]), 1e-3)
            assert relocated_grids.sub[35] == pytest.approx(np.array([0.5, 0.3]), 1e-3)

        def test__8_points_with_border_as_circle__points_go_to_circle_edge(self):

            grid = np.array([[1.0, 0.0], [0.0, 1.0], [-1.0, 0.0], [0.0 ,-1.0],
                             [0.7071, 0.7071], [0.7071, -0.7071],
                             [-0.7071, 0.7071], [-0.7071, -0.7071],
                             [10., 10.], [10., -10.], [-10., 10], [-10., -10.]])
            data_grids = grids.DataGrids(regular=grid, sub=grid, blurring=None)

            border_pixels = np.array([0, 1, 2, 3, 4, 5, 6, 7])

            border = grids.RegularGridBorder(border_pixels)

            relocated_grids = border.relocated_grids_from_grids(data_grids)

            assert relocated_grids.regular[0] == pytest.approx(grid[0], 1e-3)
            assert relocated_grids.regular[1] == pytest.approx(grid[1], 1e-3)
            assert relocated_grids.regular[2] == pytest.approx(grid[2], 1e-3)
            assert relocated_grids.regular[3] == pytest.approx(grid[3], 1e-3)
            assert relocated_grids.regular[4] == pytest.approx(grid[4], 1e-3)
            assert relocated_grids.regular[5] == pytest.approx(grid[5], 1e-3)
            assert relocated_grids.regular[6] == pytest.approx(grid[6], 1e-3)
            assert relocated_grids.regular[7] == pytest.approx(grid[7], 1e-3)
            assert relocated_grids.regular[8] == pytest.approx(np.array([0.7071, 0.7071]), 1e-3)
            assert relocated_grids.regular[9] == pytest.approx(np.array([0.7071, -0.7071]), 1e-3)
            assert relocated_grids.regular[10] == pytest.approx(np.array([-0.7071, 0.7071]), 1e-3)
            assert relocated_grids.regular[11] == pytest.approx(np.array([-0.7071, -0.7071]), 1e-3)

            assert relocated_grids.sub[0] == pytest.approx(grid[0], 1e-3)
            assert relocated_grids.sub[1] == pytest.approx(grid[1], 1e-3)
            assert relocated_grids.sub[2] == pytest.approx(grid[2], 1e-3)
            assert relocated_grids.sub[3] == pytest.approx(grid[3], 1e-3)
            assert relocated_grids.sub[4] == pytest.approx(grid[4], 1e-3)
            assert relocated_grids.sub[5] == pytest.approx(grid[5], 1e-3)
            assert relocated_grids.sub[6] == pytest.approx(grid[6], 1e-3)
            assert relocated_grids.sub[7] == pytest.approx(grid[7], 1e-3)
            assert relocated_grids.sub[8] == pytest.approx(np.array([0.7071, 0.7071]), 1e-3)
            assert relocated_grids.sub[9] == pytest.approx(np.array([0.7071, -0.7071]), 1e-3)
            assert relocated_grids.sub[10] == pytest.approx(np.array([-0.7071, 0.7071]), 1e-3)
            assert relocated_grids.sub[11] == pytest.approx(np.array([-0.7071, -0.7071]), 1e-3)

        def test__same_as_above_but_ensure_positive_origin_moves_points(self):

            grid = np.array([[2.0, 1.0], [1.0, 2.0], [0.0, 1.0], [1.0, 0.0],
                             [1.0 + 0.7071, 1.0 + 0.7071], [1.0 + 0.7071, 1.0 - 0.7071],
                             [1.0 - 0.7071, 1.0 + 0.7071], [1.0 - 0.7071, 1.0 - 0.7071],
                             [11., 11.], [11., -9.], [-9., 11], [-9., -9.]])
            data_grids = grids.DataGrids(regular=grid, sub=grid, blurring=None)

            border_pixels = np.array([0, 1, 2, 3, 4, 5, 6, 7])

            border = grids.RegularGridBorder(border_pixels)

            relocated_grids = border.relocated_grids_from_grids(data_grids)

            assert relocated_grids.regular[0] == pytest.approx(grid[0], 1e-3)
            assert relocated_grids.regular[1] == pytest.approx(grid[1], 1e-3)
            assert relocated_grids.regular[2] == pytest.approx(grid[2], 1e-3)
            assert relocated_grids.regular[3] == pytest.approx(grid[3], 1e-3)
            assert relocated_grids.regular[4] == pytest.approx(grid[4], 1e-3)
            assert relocated_grids.regular[5] == pytest.approx(grid[5], 1e-3)
            assert relocated_grids.regular[6] == pytest.approx(grid[6], 1e-3)
            assert relocated_grids.regular[7] == pytest.approx(grid[7], 1e-3)
            assert relocated_grids.regular[8] == pytest.approx(np.array([1.0 + 0.7071, 1.0 + 0.7071]), 1e-3)
            assert relocated_grids.regular[9] == pytest.approx(np.array([1.0 + 0.7071, 1.0 - 0.7071]), 1e-3)
            assert relocated_grids.regular[10] == pytest.approx(np.array([1.0 - 0.7071, 1.0 + 0.7071]), 1e-3)
            assert relocated_grids.regular[11] == pytest.approx(np.array([1.0 - 0.7071, 1.0 - 0.7071]), 1e-3)

            assert relocated_grids.sub[0] == pytest.approx(grid[0], 1e-3)
            assert relocated_grids.sub[1] == pytest.approx(grid[1], 1e-3)
            assert relocated_grids.sub[2] == pytest.approx(grid[2], 1e-3)
            assert relocated_grids.sub[3] == pytest.approx(grid[3], 1e-3)
            assert relocated_grids.sub[4] == pytest.approx(grid[4], 1e-3)
            assert relocated_grids.sub[5] == pytest.approx(grid[5], 1e-3)
            assert relocated_grids.sub[6] == pytest.approx(grid[6], 1e-3)
            assert relocated_grids.sub[7] == pytest.approx(grid[7], 1e-3)
            assert relocated_grids.sub[8] == pytest.approx(np.array([1.0 + 0.7071, 1.0 + 0.7071]), 1e-3)
            assert relocated_grids.sub[9] == pytest.approx(np.array([1.0 + 0.7071, 1.0 - 0.7071]), 1e-3)
            assert relocated_grids.sub[10] == pytest.approx(np.array([1.0 - 0.7071, 1.0 + 0.7071]), 1e-3)
            assert relocated_grids.sub[11] == pytest.approx(np.array([1.0 - 0.7071, 1.0 - 0.7071]), 1e-3)

        def test__same_as_above_but_ensure_negative_origin_moves_points(self):

            grid = np.array([[0.0, -1.0], [-1.0, 0.0], [-2.0, -1.0], [-1.0 ,-2.0],
                             [-1.0 + 0.7071, -1.0 + 0.7071], [-1.0 + 0.7071, -1.0 - 0.7071],
                             [-1.0 - 0.7071, -1.0 + 0.7071], [-1.0 - 0.7071, -1.0 - 0.7071],
                             [9., 9.], [9., -11.], [-11., 9], [-11., -11.]])
            data_grids = grids.DataGrids(regular=grid, sub=grid, blurring=None)
            border_pixels = np.array([0, 1, 2, 3, 4, 5, 6, 7])

            border = grids.RegularGridBorder(border_pixels)

            relocated_grids = border.relocated_grids_from_grids(data_grids)

            assert relocated_grids.regular[0] == pytest.approx(grid[0], 1e-3)
            assert relocated_grids.regular[1] == pytest.approx(grid[1], 1e-3)
            assert relocated_grids.regular[2] == pytest.approx(grid[2], 1e-3)
            assert relocated_grids.regular[3] == pytest.approx(grid[3], 1e-3)
            assert relocated_grids.regular[4] == pytest.approx(grid[4], 1e-3)
            assert relocated_grids.regular[5] == pytest.approx(grid[5], 1e-3)
            assert relocated_grids.regular[6] == pytest.approx(grid[6], 1e-3)
            assert relocated_grids.regular[7] == pytest.approx(grid[7], 1e-3)
            assert relocated_grids.regular[8] == pytest.approx(np.array([-1.0 + 0.7071, -1.0 + 0.7071]), 1e-3)
            assert relocated_grids.regular[9] == pytest.approx(np.array([-1.0 + 0.7071, -1.0 - 0.7071]), 1e-3)
            assert relocated_grids.regular[10] == pytest.approx(np.array([-1.0 - 0.7071, -1.0 + 0.7071]), 1e-3)
            assert relocated_grids.regular[11] == pytest.approx(np.array([-1.0 - 0.7071, -1.0 - 0.7071]), 1e-3)

            assert relocated_grids.sub[0] == pytest.approx(grid[0], 1e-3)
            assert relocated_grids.sub[1] == pytest.approx(grid[1], 1e-3)
            assert relocated_grids.sub[2] == pytest.approx(grid[2], 1e-3)
            assert relocated_grids.sub[3] == pytest.approx(grid[3], 1e-3)
            assert relocated_grids.sub[4] == pytest.approx(grid[4], 1e-3)
            assert relocated_grids.sub[5] == pytest.approx(grid[5], 1e-3)
            assert relocated_grids.sub[6] == pytest.approx(grid[6], 1e-3)
            assert relocated_grids.sub[7] == pytest.approx(grid[7], 1e-3)
            assert relocated_grids.sub[8] == pytest.approx(np.array([-1.0 + 0.7071, -1.0 + 0.7071]), 1e-3)
            assert relocated_grids.sub[9] == pytest.approx(np.array([-1.0 + 0.7071, -1.0 - 0.7071]), 1e-3)
            assert relocated_grids.sub[10] == pytest.approx(np.array([-1.0 - 0.7071, -1.0 + 0.7071]), 1e-3)
            assert relocated_grids.sub[11] == pytest.approx(np.array([-1.0 - 0.7071, -1.0 - 0.7071]), 1e-3)

        def test__point_is_inside_border_but_further_than_minimum_border_point_radii__does_not_relocate(self):

            grid = np.array([[1.0, 0.0], [0.0, 1.0], [-1.0, 0.0], [0.0, -0.9],
                             [0.7071, 0.7071], [0.7071, -0.7071],
                             [-0.7071, 0.7071], [-0.7071, -0.7071],
                             [0.02, 0.95]])

            data_grids = grids.DataGrids(regular=grid, sub=grid, blurring=None)

            border_pixels = np.array([0, 1, 2, 3, 4, 5, 6, 7])

            border = grids.RegularGridBorder(border_pixels)

            relocated_grids = border.relocated_grids_from_grids(data_grids)

            assert relocated_grids.regular[8] == pytest.approx(np.array([0.02, 0.95]), 1e-4)
            assert relocated_grids.sub[8] == pytest.approx(np.array([0.02, 0.95]), 1e-4)

        def test__inside_border_no_relocations__also_include_pix_grid(self):

            thetas = np.linspace(0.0, 2.0 * np.pi, 32)
            grid_circle = list(map(lambda x: (np.cos(x), np.sin(x)), thetas))
            grid = grid_circle
            grid.append(np.array([0.1, 0.0]))
            grid.append(np.array([-0.2, -0.3]))
            grid.append(np.array([0.5, 0.4]))
            grid.append(np.array([0.7, -0.1]))
            image_grid = np.asarray(grid)
            sub_grid = np.asarray(grid)
            sub_grid[35 ,0] = 0.5
            sub_grid[35 ,1] = 0.3
            data_grids = grids.DataGrids(regular=image_grid, sub=sub_grid, blurring=None, pix=image_grid)

            border = grids.RegularGridBorder(arr=np.arange(32))
            relocated_grids = border.relocated_grids_from_grids(data_grids)

            assert relocated_grids.regular[0:32] == pytest.approx(np.asarray(grid_circle)[0:32], 1e-3)
            assert relocated_grids.regular[32] == pytest.approx(np.array([0.1, 0.0]), 1e-3)
            assert relocated_grids.regular[33] == pytest.approx(np.array([-0.2, -0.3]), 1e-3)
            assert relocated_grids.regular[34] == pytest.approx(np.array([0.5, 0.4]), 1e-3)
            assert relocated_grids.regular[35] == pytest.approx(np.array([0.7, -0.1]), 1e-3)

            assert relocated_grids.sub[0:32] == pytest.approx(np.asarray(grid_circle)[0:32], 1e-3)
            assert relocated_grids.sub[32] == pytest.approx(np.array([0.1, 0.0]), 1e-3)
            assert relocated_grids.sub[33] == pytest.approx(np.array([-0.2, -0.3]), 1e-3)
            assert relocated_grids.sub[34] == pytest.approx(np.array([0.5, 0.4]), 1e-3)
            assert relocated_grids.sub[35] == pytest.approx(np.array([0.5, 0.3]), 1e-3)
            
            assert relocated_grids.pix[0:32] == pytest.approx(np.asarray(grid_circle)[0:32], 1e-3)
            assert relocated_grids.pix[32] == pytest.approx(np.array([0.1, 0.0]), 1e-3)
            assert relocated_grids.pix[33] == pytest.approx(np.array([-0.2, -0.3]), 1e-3)
            assert relocated_grids.pix[34] == pytest.approx(np.array([0.5, 0.4]), 1e-3)
            assert relocated_grids.pix[35] == pytest.approx(np.array([0.7, -0.1]), 1e-3)


import numpy as np

from autolens.data.imaging import image as im
from autolens.data.array import mask as msk
from autolens.data.fitting import fitting_data
from autolens.model.inversion import convolution as inversion_convolution


class LensingImage(fitting_data.FittingImage):

    def __new__(cls, image, mask, sub_grid_size=2, image_psf_shape=None, mapping_matrix_psf_shape=None, positions=None):
        return np.array(mask.map_2d_array_to_masked_1d_array(image)).view(cls)

    def __init__(self, image, mask, sub_grid_size=2, image_psf_shape=None, mapping_matrix_psf_shape=None,
                 positions=None):
        """
        The lensing datas_ is the collection of datas (regular, noise-maps, PSF), a masks, grids, convolvers and other \
        utilities that are used for modeling and fitting an datas_ of a strong lens.

        Whilst the datas_ datas is initially loaded in 2D, for the lensing datas_ the masked-datas_ (and noise-maps) \
        are reduced to 1D arrays for faster calculations.

        Parameters
        ----------
        image: im.Image
            The original datas_ datas in 2D.
        mask: msk.Mask
            The 2D masks that is applied to the datas_.
        sub_grid_size : int
            The size of the sub-grid used for each lensing SubGrid. E.g. a value of 2 grids each datas_-pixel on a 2x2 \
            sub-grid.
        image_psf_shape : (int, int)
            The shape of the PSF used for convolving model regular generated using analytic light profiles. A smaller \
            shape will trim the PSF relative to the input datas_ PSF, giving a faster analysis run-time.
        mapping_matrix_psf_shape : (int, int)
            The shape of the PSF used for convolving the inversion mapping matrix. A smaller \
            shape will trim the PSF relative to the input datas_ PSF, giving a faster analysis run-time.
        positions : [[]]
            Lists of datas_-pixel coordinates (arc-seconds) that mappers close to one another in the source-plane(s), used \
            to speed up the non-linear sampling.
        """
        super().__init__(image=image, mask=mask, sub_grid_size=sub_grid_size, image_psf_shape=image_psf_shape)

        if mapping_matrix_psf_shape is None:
            mapping_matrix_psf_shape = self.image.psf.shape

        self.convolver_mapping_matrix = inversion_convolution.ConvolverMappingMatrix(self.mask,
                      self.image.psf.resized_scaled_array_from_array(mapping_matrix_psf_shape))

        self.positions = positions


    def __array_finalize__(self, obj):
        super(LensingImage, self).__array_finalize__(obj)
        if isinstance(obj, LensingImage):
            self.image = obj.image
            self.mask = obj.mask
            self.noise_map_ = obj.noise_map_
            self.background_noise_map_ = obj.background_noise_map_
            self.poisson_noise_map_ = obj.poisson_noise_map_
            self.exposure_time_map_ = obj.exposure_time_map_
            self.background_sky_map_ = obj.background_sky_map_
            self.convolver_image = obj.convolver_image
            self.convolver_mapping_matrix = obj.convolver_mapping_matrix
            self.sub_grid_size = obj.sub_grid_size
            self.grids = obj.grids
            self.padded_grids = obj.padded_grids
            self.border = obj.border
            self.positions = obj.positions


class LensingHyperImage(fitting_data.FittingHyperImage):

    def __new__(cls, image, mask, hyper_model_image, hyper_galaxy_images, hyper_minimum_values, sub_grid_size=2,
                image_psf_shape=None, mapping_matrix_psf_shape=None, positions=None):
        return np.array(mask.map_2d_array_to_masked_1d_array(image)).view(cls)

    def __init__(self, image, mask, hyper_model_image, hyper_galaxy_images, hyper_minimum_values, sub_grid_size=2,
                 image_psf_shape=None, mapping_matrix_psf_shape=None, positions=None):
        """
        The lensing datas_ is the collection of datas (regular, noise-maps, PSF), a masks, grids, convolvers and other \
        utilities that are used for modeling and fitting an datas_ of a strong lens.

        Whilst the datas_ datas is initially loaded in 2D, for the lensing datas_ the masked-datas_ (and noise-maps) \
        are reduced to 1D arrays for faster calculations.

        Parameters
        ----------
        image: im.Image
            The original datas_ datas in 2D.
        mask: msk.Mask
            The 2D masks that is applied to the datas_.
        sub_grid_size : int
            The size of the sub-grid used for each lensing SubGrid. E.g. a value of 2 grids each datas_-pixel on a 2x2 \
            sub-grid.
        image_psf_shape : (int, int)
            The shape of the PSF used for convolving model regular generated using analytic light profiles. A smaller \
            shape will trim the PSF relative to the input datas_ PSF, giving a faster analysis run-time.
        mapping_matrix_psf_shape : (int, int)
            The shape of the PSF used for convolving the inversion mapping matrix. A smaller \
            shape will trim the PSF relative to the input datas_ PSF, giving a faster analysis run-time.
        positions : [[]]
            Lists of datas_-pixel coordinates (arc-seconds) that mappers close to one another in the source-plane(s), used \
            to speed up the non-linear sampling.
        """
        super().__init__(image=image, mask=mask, hyper_model_image=hyper_model_image,
                         hyper_galaxy_images=hyper_galaxy_images, hyper_minimum_values=hyper_minimum_values,
                         sub_grid_size=sub_grid_size, image_psf_shape=image_psf_shape)

        if mapping_matrix_psf_shape is None:
            mapping_matrix_psf_shape = self.image.psf.shape

        self.convolver_mapping_matrix = inversion_convolution.ConvolverMappingMatrix(self.mask,
                      self.image.psf.resized_scaled_array_from_array(mapping_matrix_psf_shape))

        self.positions = positions

    def __array_finalize__(self, obj):
        super(LensingHyperImage, self).__array_finalize__(obj)
        if isinstance(obj, LensingHyperImage):
            self.image = obj.image
            self.mask = obj.mask
            self.convolver_image = obj.convolver_image
            self.convolver_mapping_matrix = obj.convolver_mapping_matrix
            self.grids = obj.grids
            self.sub_grid_size = obj.sub_grid_size
            self.padded_grids = obj.padded_grids
            self.border = obj.border
            self.positions = obj.positions
            self.hyper_model_image = obj.hyper_model_image
            self.hyper_galaxy_images = obj.hyper_galaxy_images
            self.hyper_minimum_values = obj.hyper_minimum_values


from matplotlib import pyplot as plt
import numpy as np
import itertools

from autolens import exc
from autolens.data.array.plotters import plotter_util


def plot_grid(grid, axis_limits=None, points=None, as_subplot=False,
              units='arcsec', kpc_per_arcsec=None,
              figsize=(12, 8), pointsize=5, pointcolor='k', xyticksize=16,
              title='Grid', titlesize=16, xlabelsize=16, ylabelsize=16,
              output_path=None, output_format='show', output_filename='grid'):

    plotter_util.setup_figure(figsize=figsize, as_subplot=as_subplot)
    grid = convert_grid_units(grid_arc_seconds=grid, units=units, kpc_per_arcsec=kpc_per_arcsec)
    plt.scatter(y=np.asarray(grid[:, 0]), x=np.asarray(grid[:, 1]), s=pointsize, marker='.')
    plotter_util.set_title(title=title, titlesize=titlesize)
    set_xy_labels_and_ticks_in_arcsec(units, kpc_per_arcsec, grid.xticks, grid.yticks, xlabelsize, ylabelsize,
                                      xyticksize)

    set_axis_limits(axis_limits)
    plot_points(grid, points, pointcolor)

    plt.tick_params(labelsize=xyticksize)
    plotter_util.output_figure(None, as_subplot, output_path, output_filename, output_format)
    plotter_util.close_figure(as_subplot=as_subplot)

def convert_grid_units(grid_arc_seconds, units, kpc_per_arcsec):

    if units is 'arcsec' or kpc_per_arcsec is None:
        return grid_arc_seconds
    elif units is 'kpc':
        return grid_arc_seconds * kpc_per_arcsec

def set_xy_labels_and_ticks_in_arcsec(units, kpc_per_arcsec, xticks, yticks, xlabelsize, ylabelsize, xyticksize):

    if units is 'arcsec' or kpc_per_arcsec is None:

        plt.xlabel('x (arcsec)', fontsize=xlabelsize)
        plt.ylabel('y (arcsec)', fontsize=ylabelsize)

    elif units is 'kpc':

        plt.xlabel('x (kpc)', fontsize=xlabelsize)
        plt.ylabel('y (kpc)', fontsize=ylabelsize)

    else:
        raise exc.PlottingException('The units supplied to the plotted are not a valid string (must be pixels | '
                                     'arcsec | kpc)')

    plt.tick_params(labelsize=xyticksize)

def set_axis_limits(axis_limits):

    if axis_limits is not None:
        plt.axis(axis_limits)

def plot_points(grid, points, pointcolor):

    if points is not None:

        if pointcolor is None:

            point_colors = itertools.cycle(["y", "r", "k", "g", "m"])
            for point_set in points:
                plt.scatter(y=np.asarray(grid[point_set, 0]),
                            x=np.asarray(grid[point_set, 1]), s=8, color=next(point_colors))

        else:

            for point_set in points:
                plt.scatter(y=np.asarray(grid[point_set, 0]),
                            x=np.asarray(grid[point_set, 1]), s=8, color=pointcolor)



import os
import shutil

import numpy as np

from autofit import conf
from autolens.data.imaging import image as im
from autolens.data.array.util import array_util
from autolens.data.array import grids, scaled_array
from autolens.lensing import ray_tracing

dirpath = os.path.dirname(os.path.realpath(__file__))

def reset_paths(data_name, pipeline_name, output_path):

    conf.instance.output_path = output_path

    try:
        shutil.rmtree(dirpath + '/data' + data_name)
    except FileNotFoundError:
        pass

    try:
        shutil.rmtree(output_path + '/' + pipeline_name)
    except FileNotFoundError:
        pass

def simulate_integration_image(data_name, pixel_scale, lens_galaxies, source_galaxies, target_signal_to_noise):
    output_path = "{}/data/".format(os.path.dirname(os.path.realpath(__file__))) + data_name + '/'
    psf_shape = (11, 11)
    image_shape = (150, 150)

    psf = im.PSF.simulate_as_gaussian(shape=psf_shape, pixel_scale=pixel_scale, sigma=pixel_scale)

    image_grids = grids.DataGrids.grids_for_simulation(shape=image_shape, pixel_scale=pixel_scale,
                                                       sub_grid_size=1, psf_shape=psf_shape)

    image_shape = image_grids.regular.padded_shape

    if not source_galaxies:

        tracer = ray_tracing.TracerImagePlane(lens_galaxies=lens_galaxies, image_plane_grids=[image_grids])

    elif source_galaxies:

        tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=lens_galaxies, source_galaxies=source_galaxies,
                                                     image_plane_grids=[image_grids])

    ### Setup as a simulated image_coords and output as a fits for an lensing ###

    sim_image = im.Image.simulate_to_target_signal_to_noise(array=tracer.image_plane_images_for_simulation[0],
                                                            pixel_scale=pixel_scale,
                                                            target_signal_to_noise=target_signal_to_noise,
                                                            exposure_time_map=np.ones(image_shape),
                                                            background_sky_map=10.0 * np.ones(image_shape),
                                                            psf=psf, seed=1)

    if os.path.exists(output_path) == False:
        os.makedirs(output_path)

    array_util.numpy_array_to_fits(sim_image, path=output_path + 'regular.fits')
    array_util.numpy_array_to_fits(sim_image.noise_map, path=output_path + 'noise_map_.fits')
    array_util.numpy_array_to_fits(psf, path=output_path + '/psf.fits')
    array_util.numpy_array_to_fits(sim_image.exposure_time_map, path=output_path + 'exposure_map.fits')


def load_image(data_name, pixel_scale):
    data_dir = "{}/data/{}".format(dirpath, data_name)

    data = scaled_array.ScaledSquarePixelArray.from_fits_with_pixel_scale(file_path=data_dir + '/regular.fits', hdu=0,
                                                                          pixel_scale=pixel_scale)
    noise = scaled_array.ScaledSquarePixelArray.from_fits_with_pixel_scale(file_path=data_dir + '/noise_map_.fits', hdu=0,
                                                                           pixel_scale=pixel_scale)
    psf = im.PSF.from_fits_with_scale(file_path=data_dir + '/psf.fits', hdu=0, pixel_scale=pixel_scale)

    return im.Image(array=data, pixel_scale=pixel_scale, psf=psf, noise_map=noise)



from autolens.data.array import grids
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
import numpy as np
import pytest


# Arc second coordinate grid is:.

# [[[-2.0, -2.0], [-2.0, -1.0], [-2.0, 0.0], [-2.0, 1.0], [-2.0, 2.0]],
# [[[-1.0, -2.0], [-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0], [-1.0, 2.0]],
# [[[ 0.0, -2.0], [ 0.0, -1.0], [ 0.0, 0.0], [ 0.0, 1.0], [ 0.0, 2.0]],
# [[[ 1.0, -2.0], [ 1.0, -1.0], [ 1.0, 0.0], [ 1.0, 1.0], [ 1.0, 2.0]],
# [[[ 2.0, -2.0], [ 2.0, -1.0], [ 2.0, 0.0], [ 2.0, 1.0], [ 2.0, 2.0]],


def test__centre_light_profile_on_grid_coordinate__peak_flux_is_correct_index():

    image_grid = grids.RegularGrid.from_shape_and_pixel_scale(shape=(5, 5), pixel_scale=1.0)

    sersic = lp.SphericalSersic(centre=(2.0, -2.0))
    image_1d = sersic.intensities_from_grid(grid=image_grid)
    image_2d = image_grid.map_to_2d(array_1d=image_1d)

    assert image_1d.argmax() == 0
    assert np.unravel_index(image_2d.argmax(), image_2d.shape) == (0, 0)

    sersic = lp.SphericalSersic(centre=(2.0, 2.0))
    image_1d = sersic.intensities_from_grid(grid=image_grid)
    image_2d = image_grid.map_to_2d(array_1d=image_1d)

    assert image_1d.argmax() == 4
    assert np.unravel_index(image_2d.argmax(), image_2d.shape) == (0, 4)

    sersic = lp.SphericalSersic(centre=(-2.0, -2.0))
    image_1d = sersic.intensities_from_grid(grid=image_grid)
    image_2d = image_grid.map_to_2d(array_1d=image_1d)

    assert image_1d.argmax() == 20
    assert np.unravel_index(image_2d.argmax(), image_2d.shape) == (4, 0)

    sersic = lp.SphericalSersic(centre=(-2.0, 2.0))
    image_1d = sersic.intensities_from_grid(grid=image_grid)
    image_2d = image_grid.map_to_2d(array_1d=image_1d)

    assert image_1d.argmax() == 24
    assert np.unravel_index(image_2d.argmax(), image_2d.shape) == (4, 4)

def test__centre_mass_profile_on_grid_coordinate__peak_density_is_correct_index():

    image_grid = grids.RegularGrid.from_shape_and_pixel_scale(shape=(5, 5), pixel_scale=1.0)

    sis = mp.SphericalIsothermal(centre=(2.0, -2.0))
    density_1d = sis.surface_density_from_grid(grid=image_grid)
    density_2d = image_grid.map_to_2d(array_1d=density_1d)

    assert density_1d.argmax() == 0
    assert np.unravel_index(density_2d.argmax(), density_2d.shape) == (0, 0)

    sis = mp.SphericalIsothermal(centre=(2.0, 2.0))
    density_1d = sis.surface_density_from_grid(grid=image_grid)
    density_2d = image_grid.map_to_2d(array_1d=density_1d)

    assert density_1d.argmax() == 4
    assert np.unravel_index(density_2d.argmax(), density_2d.shape) == (0, 4)

    sis = mp.SphericalIsothermal(centre=(-2.0, -2.0))
    density_1d = sis.surface_density_from_grid(grid=image_grid)
    density_2d = image_grid.map_to_2d(array_1d=density_1d)

    assert density_1d.argmax() == 20
    assert np.unravel_index(density_2d.argmax(), density_2d.shape) == (4, 0)

    sis =  mp.SphericalIsothermal(centre=(-2.0, 2.0))
    density_1d = sis.surface_density_from_grid(grid=image_grid)
    density_2d = image_grid.map_to_2d(array_1d=density_1d)

    assert density_1d.argmax() == 24
    assert np.unravel_index(density_2d.argmax(), density_2d.shape) == (4, 4)

def test__same_as_above__but_grid_is_padded_to_7x7_for_simulation():

    data_grids = grids.DataGrids.grids_for_simulation(shape=(5, 5), pixel_scale=1.0, psf_shape=(3, 3))

    sersic = lp.SphericalSersic(centre=(2.0, -2.0))
    image_1d = sersic.intensities_from_grid(grid=data_grids.regular)
    assert image_1d.argmax() == 8
    image_2d = data_grids.regular.map_to_2d(padded_array_1d=image_1d)
    assert np.unravel_index(image_2d.argmax(), image_2d.shape) == (0, 0)
    image_2d = data_grids.regular.map_to_2d_keep_padded(padded_array_1d=image_1d)
    assert np.unravel_index(image_2d.argmax(), image_2d.shape) == (1, 1)

    sersic = lp.SphericalSersic(centre=(2.0, 2.0))
    image_1d = sersic.intensities_from_grid(grid=data_grids.regular)
    assert image_1d.argmax() == 12
    image_2d = data_grids.regular.map_to_2d(padded_array_1d=image_1d)
    assert np.unravel_index(image_2d.argmax(), image_2d.shape) == (0, 4)
    image_2d = data_grids.regular.map_to_2d_keep_padded(padded_array_1d=image_1d)
    assert np.unravel_index(image_2d.argmax(), image_2d.shape) == (1, 5)

    sersic = lp.SphericalSersic(centre=(-2.0, -2.0))
    image_1d = sersic.intensities_from_grid(grid=data_grids.regular)
    assert image_1d.argmax() == 36
    image_2d = data_grids.regular.map_to_2d(padded_array_1d=image_1d)
    assert np.unravel_index(image_2d.argmax(), image_2d.shape) == (4, 0)
    image_2d = data_grids.regular.map_to_2d_keep_padded(padded_array_1d=image_1d)
    assert np.unravel_index(image_2d.argmax(), image_2d.shape) == (5, 1)

    sersic = lp.SphericalSersic(centre=(-2.0, 2.0))
    image_1d = sersic.intensities_from_grid(grid=data_grids.regular)
    assert image_1d.argmax() == 40
    image_2d = data_grids.regular.map_to_2d(padded_array_1d=image_1d)
    assert np.unravel_index(image_2d.argmax(), image_2d.shape) == (4, 4)
    image_2d = data_grids.regular.map_to_2d_keep_padded(padded_array_1d=image_1d)
    assert np.unravel_index(image_2d.argmax(), image_2d.shape) == (5, 5)

def test__deflection_angles():

    image_grid = grids.RegularGrid.from_shape_and_pixel_scale(shape=(5, 5), pixel_scale=1.0)

    sis = mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)
    deflections_1d = sis.deflections_from_grid(grid=image_grid)
    deflections_x_2d = image_grid.map_to_2d(array_1d=deflections_1d[:,1])

    assert deflections_x_2d[0,0] == pytest.approx(-1.0*deflections_x_2d[0,4], 1e-2)
    assert deflections_x_2d[1,1] == pytest.approx(-1.0*deflections_x_2d[1,3], 1e-2)
    assert deflections_x_2d[3,1] == pytest.approx(-1.0*deflections_x_2d[3,3], 1e-2)
    assert deflections_x_2d[4,0] == pytest.approx(-1.0*deflections_x_2d[4,4], 1e-2)
    assert deflections_x_2d[0,2] == pytest.approx(deflections_x_2d[4,2], 1e-2)

    sis = mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)
    deflections_1d = sis.deflections_from_grid(grid=image_grid)
    deflections_y_2d = image_grid.map_to_2d(array_1d=deflections_1d[:,0])

    assert deflections_y_2d[0,0] == pytest.approx(-1.0*deflections_y_2d[4,0], 1e-2)
    assert deflections_y_2d[1,1] == pytest.approx(-1.0*deflections_y_2d[3,1], 1e-2)
    assert deflections_y_2d[1,3] == pytest.approx(-1.0*deflections_y_2d[3,3], 1e-2)
    assert deflections_y_2d[0,4] == pytest.approx(-1.0*deflections_y_2d[4,4], 1e-2)
    assert deflections_y_2d[2,0] == pytest.approx(deflections_y_2d[2,4], 1e-2)

# def test__move_source_galaxy_around_source_plane__peak_follows_source_direction():
#
#     image_grid = masks.RegularGrid.from_shape_and_pixel_scale(shape=(5, 5), pixel_scales=1.0)
#     sis = mp.SphericalIsothermal(origin=(0.0, 0.0), einstein_radius=1.0)
#     sersic = lp.SphericalSersic(origin=(1.0, 0.0))
#
#     deflections = sis.deflections_from_grid(grid=image_grid)
#     source_grid = np.subtract(image_grid, deflections)
#     print(image_grid[22])
#     print(deflections[22])
#     print(source_grid[22])
#     stop
#     source_image = sersic.intensities_from_grid(grid=source_grid)
#     print(source_image.argmax())
#
#     data_grids = masks.DataGrids.from_shape_and_pixel_scale(shape=(5, 5), pixel_scales=1.0)
#     lens_galaxy = g.Galaxy(mass=sis)
#     source_galaxy = g.Galaxy(light=sersic)
#     tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
#                                                  image_plane_grids=data_grids)
#
#     print(source_grid)
#     print(tracer.source_plane.grids.regular)
#     print(np.subtract(source_grid, tracer.source_plane.grids.regular))
#     print(tracer.regular_plane_image)


from autofit import conf
from autofit.core import non_linear as nl
from autofit.core import model_mapper as mm
from autolens.pipeline import phase as ph
from autolens.model.galaxy import galaxy_model as gm
from autolens.data.imaging import image as im
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.data.imaging.plotters import imaging_plotters
from autolens.lensing.plotters import lensing_fitting_plotters

import os

# So, we've learnt that if our parameter space is too complex, our non-linear search might fail to find the global
# maximum solution. However, we also learnt how to ensure this doesn't happen, by:

# 1) Tuning our priors to the strong lens we're fitting.
# 2) Making our lens model less complex.
# 3) Searching non-linear parameter space for longer.

# However, each of the above approaches has disadvantages. The more we tune our priors, the less we can generalize our
# analysis to a different strong lens. The less complex we make our model, the less realistic it is. And if we rely too
# much on searching parameter space for longer, we could end up with phase's that take days, weeks or months to run.

# In this exercise, we're going to combine these 3 approaches, so that we can fit_normal complex and realistic lens
# models in a way that that can be generalized to many different strong lenses. To do this, we'll run 2 phases, and
# link the lens model inferred in the first phase to the priors of the second phase's lens model.

# Our first phase will make the same light-traces-mass assumption we made in the previous tutorial. We saw that this
# gives a reasonable lens model. However, we'll make a couple of extra simplifying assumptions, to really try and
# bring our lens model complexity down and get the non-linear search running fast.

# The model we infer above will therefore be a lot less realistic. But it doesn't matter, because in the second phase
# we're going to relax these assumptions and get back our more realistic lens model. The beauty is that, by running
# the first phase, we can use its results to tune the priors of our second phase. For example:

# 1) The first phase should give us a pretty good idea of the lens model_galaxy's light and mass profiles, for example its
#    intensity, effective radius and einstein radius.

# 2) It should also give us a pretty good fit_normal to the lensed source model_galaxy. This means we'll already know where in
#    source-plane its is located and what its intensity and effective are.

#Setup the path for this run
path = '{}/../'.format(os.path.dirname(os.path.realpath(__file__)))

conf.instance = conf.Config(config_path=path+'/configs/5_linking_phases', output_path=path+"output")

# Another simulate regular function, for the same regular again.
def simulate():

    from autolens.data.array import grids
    from autolens.model.galaxy import galaxy as g
    from autolens.lensing import ray_tracing

    psf = im.PSF.simulate_as_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)
    image_plane_grids = grids.DataGrids.grids_for_simulation(shape=(130, 130), pixel_scale=0.1, psf_shape=(11, 11))

    lens_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.9, phi=45.0, intensity=0.04,
                                                             effective_radius=0.5, sersic_index=3.5),
                           mass=mp.EllipticalIsothermal(centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, einstein_radius=0.8))

    source_galaxy = g.Galaxy(light=lp.EllipticalSersic(centre=(0.0, 0.0), axis_ratio=0.5, phi=90.0, intensity=0.03,
                                                       effective_radius=0.3, sersic_index=3.0))
    tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[lens_galaxy], source_galaxies=[source_galaxy],
                                                 image_plane_grids=[image_plane_grids])

    image_simulated = im.Image.simulate(array=tracer.image_plane_image_for_simulation, pixel_scale=0.1,
                                                   exposure_time=300.0, psf=psf, background_sky_level=0.1, add_noise=True)

    return image_simulated

# Simulate the regular and set it up.
image = simulate()
imaging_plotters.plot_image_subplot(image=image)

# Lets use the same LightTracesMass Phase that we did previously, but we'll make it slightly less complex then before.

class LightTracesMassPhase(ph.LensSourcePlanePhase):

    def pass_priors(self, previous_results):

        # As we've eluded to before, one can look at an regular and immediately identify the origin of the lens
        # model_galaxy. It's that bright blob of light surrounded by the lensed source model_galaxy! Given that we know we're going
        # to make the lens model more complex in the next phase, lets take a liberal approach to the lens origin and
        # fix it to (x,y) = (0.0", 0.0").

        # You haven't actually seen a line like this one before. By setting a parameter to a number (and not a prior),
        # it is be removed rom non-linear parameter space and always fixed to that value. Pretty neat, huh?
        self.lens_galaxies.lens.light.centre_0 = 0.0
        self.lens_galaxies.lens.light.centre_1 = 0.0
        self.lens_galaxies.lens.mass.centre_0 = 0.0
        self.lens_galaxies.lens.mass.centre_1 = 0.0

        # Now, you might be thinking, doesn't this prevent our phase from generalizing to other strong lenses?
        # What if the origin of their lens model_galaxy isn't at (0.0", 0.0")?

        # Well, this is true if our datas reduction centres the lens model_galaxy somewhere else. But we get to choose where
        # we origin it when we make the regular. Therefore, I'd recommend you always origin the lens model_galaxy at the same
        # location, and (0.0", 0.0") seems the best choice!

        # We also discussed that the Sersic index of most lens galaxies is around 4. Lets be liberal and fix it to
        # 4 this time.

        self.lens_galaxies.lens.light.sersic_index = 4.0

# Now lets create the phase. You'll notice two new inputs to the lens model_galaxy's model - 'align_axis_ratios' and
# 'align_orientations'. These functions align the axis_ratio and phi values of all of the lens model_galaxy's profiles (in
# this case, its light profile and mass profile). We did this in the previous phase using the 'pass_priors' function,
# but given that this is a fairly common thing to do GalaxyModel has these coninience methods to available.

# There is also an 'align_centres' method, but because we are fixing all centres to floats we have ommitted it.
phase_1 = LightTracesMassPhase(lens_galaxies=dict(lens=gm.GalaxyModel(light=lp.EllipticalSersic,
                                                                       mass=mp.EllipticalIsothermal,
                                                   align_axis_ratios=True, align_orientations=True)),
                                source_galaxies=dict(source=gm.GalaxyModel(light=lp.EllipticalExponential)),
                                optimizer_class=nl.MultiNest, phase_name='5_linking_phase_1')

# Lets go one step further. Now we know our parameter space is less complex, maybe we can find the maximum likelihood
# with fewer MultiNest live points and a faster sampling rate?
phase_1.optimizer.n_live_points = 30
phase_1.optimizer.sampling_efficiency = 0.9

# Lets run the phase, noting that our liberal approach to reducing the lens model complexity has reduced it to just
# 11 parameters. (The results are still preloaded for you, but feel free to run it yourself, its fairly quick).
phase_1_results = phase_1.run(image=image)

# And indeed, we get a reasonably good model and fit_normal to the datas - in a much shorter space of time!
lensing_fitting_plotters.plot_fitting_subplot(fit=phase_1_results.fit)

# Now all we need to do is look at the results of phase 1 and tune our priors in phase 2 to those results. Lets
# setup a custom phase that does exactly that.

# GaussianPriors are a nice way to do this. They tell the non-linear search where in parameter space to look, in a way
# that ensures it'll look at solutions nearby. In contrast, UniformPriors puts hard limits on the boundaries of where
# the non-linear search looks, which means its more likely we'll accidently cut-out the global likelihood solution.

class CustomPriorPhase(ph.LensSourcePlanePhase):

    def pass_priors(self, previous_results):

        # What I've done here is looked at the results of phase 1, and manually specified a prior for every parameter.
        # If a parameter was fixed in the previous phase, its prior is based around the previous value. Don't worry
        # about the sigma values for now, I've chosen values that I know will ensure reasonable sampling, but we'll
        # cover this later.

        self.lens_galaxies.lens.light.centre_0 = mm.GaussianPrior(mean=0.0, sigma=0.1)
        self.lens_galaxies.lens.light.centre_1 = mm.GaussianPrior(mean=0.0, sigma=0.1)
        self.lens_galaxies.lens.light.axis_ratio = mm.GaussianPrior(mean=0.8, sigma=0.15)
        self.lens_galaxies.lens.light.phi = mm.GaussianPrior(mean=45.0, sigma=15.0)
        self.lens_galaxies.lens.light.intensity = mm.GaussianPrior(mean=0.02, sigma=0.01)
        self.lens_galaxies.lens.light.effective_radius = mm.GaussianPrior(mean=0.62, sigma=0.2)
        self.lens_galaxies.lens.light.sersic_index = mm.GaussianPrior(mean=4.0, sigma=2.0)

        self.lens_galaxies.lens.mass.centre_0 = mm.GaussianPrior(mean=0.0, sigma=0.1)
        self.lens_galaxies.lens.mass.centre_1 = mm.GaussianPrior(mean=0.0, sigma=0.1)
        self.lens_galaxies.lens.mass.axis_ratio = mm.GaussianPrior(mean=0.8, sigma=0.25)
        self.lens_galaxies.lens.mass.phi = mm.GaussianPrior(mean=45.0, sigma=30.0)
        self.lens_galaxies.lens.mass.einstein_radius = mm.GaussianPrior(mean=0.8, sigma=0.1)

        self.source_galaxies.source.light.centre_0 = mm.GaussianPrior(mean=0.0, sigma=0.1)
        self.source_galaxies.source.light.centre_0 = mm.GaussianPrior(mean=0.0, sigma=0.1)
        self.source_galaxies.source.light.axis_ratio = mm.GaussianPrior(mean=0.8, sigma=0.1)
        self.source_galaxies.source.light.phi = mm.GaussianPrior(mean=90.0, sigma=10.0)
        self.source_galaxies.source.light.intensity = mm.GaussianPrior(mean=0.14, sigma=0.05)
        self.source_galaxies.source.light.effective_radius = mm.GaussianPrior(mean=0.12, sigma=0.2)


# Lets setup and run the phase. As expected, it gives us the correct lens model. However, it does so significantly
# faster than it would had we assumed broad priros on all the parameters.
phase_2 = CustomPriorPhase(lens_galaxies=dict(lens=gm.GalaxyModel(light=lp.EllipticalSersic,
                                                                   mass=mp.EllipticalIsothermal)),
                           source_galaxies=dict(source=gm.GalaxyModel(light=lp.EllipticalExponential)),
                           optimizer_class=nl.MultiNest, phase_name='5_linking_phase_2')
phase_2.optimizer.n_live_points = 30
phase_2.optimizer.sampling_efficiency = 0.9
phase_2_results = phase_2.run(image=image)

# Look at that, the right lens model, again!
lensing_fitting_plotters.plot_fitting_subplot(fit=phase_2_results.fit)

# Our choice to link two phases together was a huge success. We managed to fit_normal a complex and realistic model,
# but were able to begin by making simplifying assumptions that eased our search of non-linear parameter space. We
# could apply phase 1 to pretty much any strong lens, and therefore get ourselves a decent lens model with which to
# tune phase 2's priors.

# You're probably thinking though, that there is one huge, giant, glaring flaw in all of this that I've not mentioned.
# Phase 2 can't be generalized to another lens - it's priors are tuned to the regular we fitted. If we had a lot of
# lenses, we'd have to write a new phase_2 for every single one. This isn't ideal, is it?

# Well, let me reassure you, that PyAutoLens has you covered. In the next set of tutorials, we'll cover 'runners'. As
# you might of guessed, a pipeline comprises a set of phases that are linked together. This allows us to start with a
# simple, easy-to-fit_normal lens model, and gradually makes it more complex. Crucially, as the pipeline runs, we 'feed' the
# results of previous phases through the pipeline, allowing us to tune our priors automatically, in a way that can be
# applied generically to any strong lens.

# However, before moving forward, there's one more thing we need to discuss - run-speed. Lens modeling is a
# computationally expensive process and keeping the run times on the order of days, or less, can be tricky. I'll give
# you a sense of how we can do that in the next exercise. Before that, think about this, what are the main factors
# driving a phase's run-time?


import numpy as np

from autolens.data.imaging import hyper_image


class TestHyperImage(object):
    class TestSkyScle(object):

        def test__scale_sky_in_image__increases_all_image_values(self):
            image = np.array([1.0, 2.0, 3.0])

            hyp = hyper_image.HyperImage(background_sky_scale=10.0)

            scaled_image = hyp.sky_scaled_image_from_image(image=image)

            assert (scaled_image == np.array([11.0, 12.0, 13.0])).all()

    class TestScaledNoise(object):

        def test__scaled_background_noise__adds_to_input_noise(self):
            noise = np.array([1.0, 2.0, 3.0])
            background_noise = np.ones((3))

            hyp = hyper_image.HyperImage(background_noise_scale=2.0)

            scaled_noise = hyp.scaled_noise_from_background_noise(noise=noise, background_noise=background_noise)

            assert (scaled_noise == np.array([3.0, 4.0, 5.0])).all()

        def test__variable_background_noise__adds_to_input_noise_different_for_each_value(self):
            noise = np.array([1.0, 2.0, 3.0])
            background_noise = np.array([5.0, 1.0, 3.0])

            hyp = hyper_image.HyperImage(background_noise_scale=3.0)

            scaled_noise = hyp.scaled_noise_from_background_noise(noise=noise, background_noise=background_noise)

            assert (scaled_noise == np.array([16.0, 5.0, 12.0])).all()



from matplotlib import pyplot as plt

from autofit import conf
from autolens.data.array.plotters import plotter_util, array_plotters
from autolens.lensing.plotters import plane_plotters


def plot_ray_tracing_subplot(tracer, mask=None, positions=None,
                             units='arcsec', figsize=None, aspect='equal',
                             cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                             cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                            titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
                             mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1.0,
                             output_path=None, output_filename='tracer_normal', output_format='show', ignore_config=True):
    """Plot the observed _tracer of an analysis, using the *Image* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    tracer : autolens.imaging.tracer_normal.Image
        Class containing the _tracer, noise-mappers and PSF that are to be plotted.
        The font size of the figure ylabel.
    output_path : str
        The path where the _tracer is output if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the _tracer is output. File formats (e.g. png, fits) output the _tracer to harddisk. 'show' displays the _tracer \
        in the python interpreter window.
    """

    plot_ray_tracing_as_subplot = conf.instance.general.get('output', 'plot_ray_tracing_as_subplot', bool)

    if plot_ray_tracing_as_subplot or ignore_config is True:

        rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=6)

        if figsize is None:
            figsize = figsize_tool

        plt.figure(figsize=figsize)
        plt.subplot(rows, columns, 1)

        plot_image_plane_image(tracer=tracer, mask=mask, positions=positions, as_subplot=True,
                               units=units, figsize=figsize, aspect=aspect,
                               cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                               linscale=linscale,
                               cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                               titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
                               mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                               output_path=output_path, output_filename='', output_format=output_format)

        plt.subplot(rows, columns, 2)

        plot_surface_density(tracer=tracer, as_subplot=True,
                             units=units, figsize=figsize, aspect=aspect,
                             cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                             linscale=linscale,
                             cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                             titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                             xyticksize=xyticksize,
                             output_path=output_path, output_filename='', output_format=output_format)

        plt.subplot(rows, columns, 3)

        plot_potential(tracer=tracer, as_subplot=True,
                       units=units, figsize=figsize, aspect=aspect,
                       cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                       linscale=linscale,
                       cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                       titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
            output_path=output_path, output_filename='', output_format=output_format)

        plt.subplot(rows, columns, 4)

        plane_plotters.plot_plane_image(plane=tracer.source_plane, as_subplot=True, positions=None, plot_grid=False,
                                        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                                        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                                        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                                        xyticksize=xyticksize,
                                        grid_pointsize=grid_pointsize,
                                        output_path=output_path, output_filename='', output_format=output_format)

        plt.subplot(rows, columns, 5)

        plot_deflections_y(tracer=tracer, as_subplot=True,
                           units=units, figsize=figsize, aspect=aspect,
                           cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                           linscale=linscale,
                           cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                           titlesize=titlesize, xlabelsize=xlabelsize,
                           ylabelsize=ylabelsize, xyticksize=xyticksize,
            output_path=output_path, output_filename='', output_format=output_format)

        plt.subplot(rows, columns, 6)

        plot_deflections_x(tracer=tracer, as_subplot=True,
                           units=units, figsize=figsize, aspect=aspect,
                           cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh,
                           linscale=linscale,
                           cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                           titlesize=titlesize, xlabelsize=xlabelsize,
                           ylabelsize=ylabelsize, xyticksize=xyticksize,
            output_path=output_path, output_filename='', output_format=output_format)

        plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                          output_format=output_format)

        plt.close()

def plot_ray_tracing_individual(tracer, output_path=None, output_format='show'):
    """Plot the observed _tracer of an analysis, using the *Image* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    tracer : autolens.imaging.tracer_normal.Image
        Class containing the _tracer, noise-mappers and PSF that are to be plotted.
        The font size of the figure ylabel.
    output_path : str
        The path where the _tracer is output if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the _tracer is output. File formats (e.g. png, fits) output the _tracer to harddisk. 'show' displays the _tracer \
        in the python interpreter window.
    """

    plot_ray_tracing_image_plane_image = conf.instance.general.get('output', 'plot_ray_tracing_image_plane_image', bool)
    plot_ray_tracing_source_plane = conf.instance.general.get('output', 'plot_ray_tracing_source_plane_image', bool)
    plot_ray_tracing_surface_density = conf.instance.general.get('output', 'plot_ray_tracing_surface_density', bool)
    plot_ray_tracing_potential = conf.instance.general.get('output', 'plot_ray_tracing_potential', bool)
    plot_ray_tracing_deflections_y = conf.instance.general.get('output', 'plot_ray_tracing_deflections_y', bool)
    plot_ray_tracing_deflections_x = conf.instance.general.get('output', 'plot_ray_tracing_deflections_x', bool)

    if plot_ray_tracing_image_plane_image:

        plot_image_plane_image(tracer=tracer, mask=None, positions=None, output_path=output_path,
                               output_format=output_format)

    if plot_ray_tracing_surface_density:

        plot_surface_density(tracer=tracer, output_path=output_path, output_format=output_format)

    if plot_ray_tracing_potential:

        plot_potential(tracer=tracer, output_path=output_path, output_format=output_format)

    if plot_ray_tracing_source_plane:

        plane_plotters.plot_plane_image(plane=tracer.source_plane, positions=None, plot_grid=False,
                                        output_path=output_path, output_filename='tracer_source_plane', output_format=output_format)

    if plot_ray_tracing_deflections_y:

        plot_deflections_y(tracer=tracer, output_path=output_path, output_format=output_format)

    if plot_ray_tracing_deflections_x:

        plot_deflections_x(tracer=tracer, output_path=output_path, output_format=output_format)


def plot_image_plane_image(tracer, image_index=0, mask=None, positions=None, as_subplot=False,
                           units='arcsec', figsize=(7, 7), aspect='equal',
                           cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                           cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                           title='Tracer Image-Plane Image', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                           mask_pointsize=10, position_pointsize=10.0,
                           output_path=None, output_format='show', output_filename='tracer_image_plane_image'):

    array_plotters.plot_array(array=tracer.image_plane_images[image_index], mask=mask, positions=positions,
                              as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=tracer.image_plane.kpc_per_arcsec_proper, figsize=figsize,
                              aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_surface_density(tracer, as_subplot=False,
                         units='arcsec', figsize=(7, 7), aspect='equal',
                         cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                         cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                         title='Tracer Surface Density', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                         output_path=None, output_format='show', output_filename='tracer_surface_density'):

    array_plotters.plot_array(array=tracer.surface_density, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=tracer.image_plane.kpc_per_arcsec_proper, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_potential(tracer, as_subplot=False,
                   units='arcsec', figsize=(7, 7), aspect='equal',
                   cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                   cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                   title='Tracer Potential', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                   output_path=None, output_format='show', output_filename='tracer_potential'):

    array_plotters.plot_array(array=tracer.potential, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=tracer.image_plane.kpc_per_arcsec_proper, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_deflections_y(tracer, as_subplot=False,
                       units='arcsec', figsize=(7, 7), aspect='equal',
                       cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                       cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                       title='Tracer Deflections (y)', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                       output_path=None, output_format='show', output_filename='tracer_deflections_y'):

    array_plotters.plot_array(array=tracer.deflections_y, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=tracer.image_plane.kpc_per_arcsec_proper, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_deflections_x(tracer, as_subplot=False,
                       units='arcsec', figsize=(7, 7), aspect='equal',
                       cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                       cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                       title='Tracer Deflections (x)', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                       output_path=None, output_format='show', output_filename='tracer_deflections_x'):

    array_plotters.plot_array(array=tracer.deflections_x, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=tracer.image_plane.kpc_per_arcsec_proper, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)



import os
import shutil
from astropy import cosmology as cosmo
import pytest
import numpy as np

from autofit import conf
from autolens.data.imaging import image as im
from autolens.data.array import grids, mask as msk, scaled_array
from autolens.lensing.plotters import lensing_fitting_plotters
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from autolens.lensing import lensing_image as li
from autolens.model.galaxy import galaxy as g
from autolens.lensing import ray_tracing
from autolens.lensing import lensing_fitting

@pytest.fixture(name='general_config')
def test_general_config():
    general_config_path = "{}/../../test_files/configs/plotting/".format(os.path.dirname(os.path.realpath(__file__)))
    conf.instance.general = conf.NamedConfig(general_config_path + "general.ini")


@pytest.fixture(name='lensing_fitting_plotter_path')
def test_lensing_fitting_plotter_setup():
    galaxy_plotter_path = "{}/../../test_files/plotting/fitting/".format(os.path.dirname(os.path.realpath(__file__)))

    if os.path.exists(galaxy_plotter_path):
        shutil.rmtree(galaxy_plotter_path)

    os.mkdir(galaxy_plotter_path)

    return galaxy_plotter_path


@pytest.fixture(name='galaxy_light')
def test_galaxy_light():
    return g.Galaxy(light=lp.EllipticalSersic(intensity=1.0), redshift=2.0)

@pytest.fixture(name='galaxy_mass')
def test_galaxy_mass():
    return g.Galaxy(mass=mp.SphericalIsothermal(einstein_radius=1.0), redshift=1.0)

@pytest.fixture(name='grids')
def test_grids():
    return grids.DataGrids.from_shape_and_pixel_scale(shape=(100, 100), pixel_scale=0.05, sub_grid_size=2)

@pytest.fixture(name='image')
def test_image():

    image = scaled_array.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=1.0)
    noise_map = im.NoiseMap(array=2.0*np.ones((3,3)), pixel_scale=1.0)
    psf = im.PSF(array=3.0*np.ones((1,1)), pixel_scale=1.0)

    return im.Image(array=image, pixel_scale=1.0, noise_map=noise_map, psf=psf)

@pytest.fixture(name='positions')
def test_positions():
    positions = [[[0.1, 0.1], [0.2, 0.2]], [[0.3, 0.3]]]
    return list(map(lambda position_set: np.asarray(position_set), positions))

@pytest.fixture(name='mask')
def test_mask():
    return msk.Mask.circular(shape=((3,3)), pixel_scale=0.1, radius_arcsec=0.1)

@pytest.fixture(name='lensing_image')
def test_lensing_image(image, mask):
    return li.LensingImage(image=image, mask=mask)

@pytest.fixture(name='fit_lens_only')
def test_fit_lens_only(lensing_image, galaxy_light):
    tracer = ray_tracing.TracerImagePlane(lens_galaxies=[galaxy_light], image_plane_grids=[lensing_image.grids],
                                          cosmology=cosmo.Planck15)
    return lensing_fitting.fit_lensing_image_with_tracer(lensing_image=lensing_image, tracer=tracer)

@pytest.fixture(name='fit_source_and_lens')
def test_fit_source_and_lens(lensing_image, galaxy_light, galaxy_mass):
    tracer = ray_tracing.TracerImageSourcePlanes(lens_galaxies=[galaxy_mass], source_galaxies=[galaxy_light],
                                               image_plane_grids=[lensing_image.grids], cosmology=cosmo.Planck15)
    return lensing_fitting.fit_lensing_image_with_tracer(lensing_image=lensing_image, tracer=tracer)

@pytest.fixture(name='hyper')
def make_hyper():
    class Hyper(object):

        def __init__(self):
            pass

    hyper = Hyper()

    hyper.hyper_model_image = np.ones((5))
    hyper.hyper_galaxy_images = [np.ones((5))]
    hyper.hyper_minimum_values = [0.2]

    hyper_galaxy = g.HyperGalaxy(contribution_factor=4.0, noise_factor=2.0, noise_power=3.0)
    hyper.hyper_galaxy = g.Galaxy(light=lp.EllipticalSersic(intensity=1.0), hyper_galaxy=hyper_galaxy)
    return hyper

@pytest.fixture(name='lensing_hyper_image')
def test_lensing_hyper_image(image, mask, hyper):

    return li.LensingHyperImage(image=image, mask=mask, hyper_model_image=hyper.hyper_model_image,
                                hyper_galaxy_images=hyper.hyper_galaxy_images,
                                hyper_minimum_values=hyper.hyper_minimum_values)

@pytest.fixture(name='fit_hyper_lens_only')
def test_fit_hyper_lens_only(lensing_hyper_image, hyper):
    tracer = ray_tracing.TracerImagePlane(lens_galaxies=[hyper.hyper_galaxy],
                                          image_plane_grids=[lensing_hyper_image.grids])
    return lensing_fitting.fit_lensing_image_with_tracer(lensing_image=lensing_hyper_image, tracer=tracer)


def test__fit_sub_plot_lens_only__output_dependent_on_config(fit_lens_only, general_config, lensing_fitting_plotter_path):

    lensing_fitting_plotters.plot_fitting_subplot(fit=fit_lens_only, should_plot_mask=True,
                                                  output_path=lensing_fitting_plotter_path, output_format='png')
    assert os.path.isfile(path=lensing_fitting_plotter_path + 'lensing_fit.png')
    os.remove(path=lensing_fitting_plotter_path + 'lensing_fit.png')

def test__fit_sub_plot_hyper_lens_only__output_dependent_on_config(fit_hyper_lens_only, general_config,
                                                                   lensing_fitting_plotter_path):

    lensing_fitting_plotters.plot_fitting_subplot(fit=fit_hyper_lens_only, should_plot_mask=True,
                                                  output_path=lensing_fitting_plotter_path,
                                                  output_filename='hyper_lensing_fit', output_format='png')
    assert os.path.isfile(path=lensing_fitting_plotter_path + 'hyper_lensing_fit.png')
    os.remove(path=lensing_fitting_plotter_path + 'hyper_lensing_fit.png')

def test__fit_sub_plot_source_and_lens__output_dependent_on_config(fit_source_and_lens, general_config,
                                                                   lensing_fitting_plotter_path):

    lensing_fitting_plotters.plot_fitting_subplot(fit=fit_source_and_lens, should_plot_mask=True,
                                                  output_path=lensing_fitting_plotter_path, output_format='png')
    assert os.path.isfile(path=lensing_fitting_plotter_path + 'lensing_fit.png')
    os.remove(path=lensing_fitting_plotter_path + 'lensing_fit.png')

def test__fit_individuals__lens_only__depedent_on_config(fit_lens_only, general_config, lensing_fitting_plotter_path):

    lensing_fitting_plotters.plot_fitting_individuals(fit=fit_lens_only, output_path=lensing_fitting_plotter_path,
                                                      output_format='png')

    assert os.path.isfile(path=lensing_fitting_plotter_path + 'fit_model_image.png')
    os.remove(path=lensing_fitting_plotter_path + 'fit_model_image.png')

    assert not os.path.isfile(path=lensing_fitting_plotter_path + 'fit_residuals.png')

    assert os.path.isfile(path=lensing_fitting_plotter_path + 'fit_chi_squareds.png')
    os.remove(path=lensing_fitting_plotter_path + 'fit_chi_squareds.png')

def test__fit_individuals__hyper_lens_only__depedent_on_config(fit_hyper_lens_only, general_config,
                                                               lensing_fitting_plotter_path):

    lensing_fitting_plotters.plot_fitting_individuals(fit=fit_hyper_lens_only, output_path=lensing_fitting_plotter_path,
                                                      output_format='png')

    assert os.path.isfile(path=lensing_fitting_plotter_path + 'fit_model_image.png')
    os.remove(path=lensing_fitting_plotter_path + 'fit_model_image.png')

    assert not os.path.isfile(path=lensing_fitting_plotter_path + 'fit_residuals.png')

    assert os.path.isfile(path=lensing_fitting_plotter_path + 'fit_chi_squareds.png')
    os.remove(path=lensing_fitting_plotter_path + 'fit_chi_squareds.png')

    assert not os.path.isfile(path=lensing_fitting_plotter_path + 'fit_scaled_chi_squareds.png')

    assert os.path.isfile(path=lensing_fitting_plotter_path + 'fit_scaled_noise_map.png')
    os.remove(path=lensing_fitting_plotter_path + 'fit_scaled_noise_map.png')

def test__fit_individuals__source_and_lens__depedent_on_config(fit_source_and_lens, general_config,
                                                               lensing_fitting_plotter_path):

    lensing_fitting_plotters.plot_fitting_individuals(fit=fit_source_and_lens, output_path=lensing_fitting_plotter_path,
                                                      output_format='png')

    assert os.path.isfile(path=lensing_fitting_plotter_path + 'fit_model_image.png')
    os.remove(path=lensing_fitting_plotter_path + 'fit_model_image.png')

    assert not os.path.isfile(path=lensing_fitting_plotter_path + 'fit_lens_plane_model_image.png')

    assert os.path.isfile(path=lensing_fitting_plotter_path + 'fit_source_plane_model_image.png')
    os.remove(path=lensing_fitting_plotter_path + 'fit_source_plane_model_image.png')

    assert not os.path.isfile(path=lensing_fitting_plotter_path + 'fit_residuals.png')

    assert os.path.isfile(path=lensing_fitting_plotter_path + 'fit_chi_squareds.png')
    os.remove(path=lensing_fitting_plotter_path + 'fit_chi_squareds.png')


import numpy as np

from autolens.data.imaging import image as im, convolution
from autolens.data.array import grids


class FittingImage(im.Image):

    def __new__(cls, image, mask, sub_grid_size=2, image_psf_shape=None):
        return np.array(mask.map_2d_array_to_masked_1d_array(image)).view(cls)

    def __init__(self, image, mask, sub_grid_size=2, image_psf_shape=None):
        """A fitting regular is the collection of data components (e.g. the regular, noise-maps, PSF, etc.) which are used \
        to fit a generate a model regular of the fitting regular and fit it to its observed regular.

        The fitting regular is masked, by converting all relevent data-components to masked 1D arrays (which are \
        syntacically followed by an underscore, e.g. noise_map_). 1D arrays are converted back to 2D masked arrays
        after the fit, using the mask's *scaled_array_from_1d_array* function.

        A fitting regular also includes a number of attributes which are used to performt the fit, including (y,x) \
        grids of coordinates, convolvers and other utilities.

        Parameters
        ----------
        image : im.Image
            The 2D observed regular and other observed quantities (noise-map, PSF, exposure-time map, etc.)
        mask: msk.Mask
            The 2D mask that is applied to regular data.
        sub_grid_size : int
            The size of the sub-grid used for computing the SubGrid (see imaging.mask.SubGrid).
        image_psf_shape : (int, int)
            The shape of the PSF used for convolving model regular generated using analytic light profiles. A smaller \
            shape will trim the PSF relative to the input regular PSF, giving a faster analysis run-time.

        Attributes
        ----------
        image : ScaledSquarePixelArray
            The 2D observed regular data (not an instance of im.Image, so does not include the other data attributes,
            which are explicitly made as new attributes of the fitting regular).
        image_ : ndarray
            The masked 1D array of the regular.
        noise_map : NoiseMap
            An array describing the RMS standard deviation error in each pixel, preferably in units of electrons per
            second.
        noise_map_ : ndarray
            The masked 1D array of the noise-map
        background_noise_map : NoiseMap or None
            An array describing the RMS standard deviation error in each pixel due to the background sky noise,
            preferably in units of electrons per second.
        background_noise_map_ : ndarray or None
            The masked 1D array of the background noise-map
        poisson_noise_map : NoiseMap or None
            An array describing the RMS standard deviation error in each pixel due to the Poisson counts of the source,
            preferably in units of electrons per second.
        poisson_noise_map_ : ndarray or None
            The masked 1D array of the poisson noise-map.
        exposure_time_map : ScaledSquarePixelArray
            An array describing the effective exposure time in each regular pixel.
        exposure_time_map_ : ndarray or None
            The masked 1D array of the exposure time-map.
        background_sky_map : ScaledSquarePixelArray
            An array describing the background sky.
        background_sky_map_ : ndarray or None
            The masked 1D array of the background sky map.
        sub_grid_size : int
            The size of the sub-grid used for computing the SubGrid (see imaging.mask.SubGrid).
        image_psf_shape : (int, int)
            The shape of the PSF used for convolving model regular generated using analytic light profiles. A smaller \
            shape will trim the PSF relative to the input regular PSF, giving a faster analysis run-time.
        convolver_image : imaging.convolution.ConvolverImage
            A convolver which convoles a 1D masked regular (using the input mask) with the 2D PSF kernel.
        grids : imaging.mask.DataGrids
            Grids of (y,x) Cartesian coordinates which map over the masked 1D observed regular's pixels (includes an \
            regular-grid, sub-grid, etc.)
        padded_grids : imaging.mask.DataGrids
            Grids of padded (y,x) Cartesian coordinates which map over the every observed regular's pixel in 1D and a \
            padded regioon to include edge's for accurate PSF convolution (includes an regular-grid, sub-grid, etc.)
        border  imaging.mask.ImagingGridsBorders
            The border of the regular-grid and sub-grid (see *ImagingGridsBorders* for their use).
        """
        super().__init__(array=image, pixel_scale=image.pixel_scale, noise_map=image.noise_map, psf=image.psf,
                         background_noise_map=image.background_noise_map, poisson_noise_map=image.poisson_noise_map,
                         exposure_time_map=image.exposure_time_map, background_sky_map=image.background_sky_map)

        self.image = image[:,:]
        self.mask = mask
        self.noise_map_ = mask.map_2d_array_to_masked_1d_array(image.noise_map)
        self.background_noise_map_ = mask.map_2d_array_to_masked_1d_array(image.background_noise_map)
        self.poisson_noise_map_ = mask.map_2d_array_to_masked_1d_array(image.poisson_noise_map)
        self.exposure_time_map_ = mask.map_2d_array_to_masked_1d_array(image.exposure_time_map)
        self.background_sky_map_ = mask.map_2d_array_to_masked_1d_array(image.background_sky_map)

        self.sub_grid_size = sub_grid_size

        if image_psf_shape is None:
            image_psf_shape = self.psf.shape

        self.convolver_image = convolution.ConvolverImage(self.mask, mask.blurring_mask_for_psf_shape(image_psf_shape),
                                                          self.psf.resized_scaled_array_from_array(image_psf_shape))

        self.grids = grids.DataGrids.grids_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                 sub_grid_size=sub_grid_size,
                                                                                 psf_shape=image_psf_shape)

        self.padded_grids = grids.DataGrids.padded_grids_from_mask_sub_grid_size_and_psf_shape(mask=mask,
                                                                                               sub_grid_size=sub_grid_size, psf_shape=image_psf_shape)

        self.border = grids.RegularGridBorder.from_mask(mask=mask)

    def __array_finalize__(self, obj):
        super(FittingImage, self).__array_finalize__(obj)
        if isinstance(obj, FittingImage):
            self.noise_map_ = obj.noise_map_
            self.background_noise_map_ = obj.background_noise_map_
            self.poisson_noise_map_ = obj.poisson_noise_map_
            self.exposure_time_map_ = obj.exposure_time_map_
            self.background_sky_map_ = obj.background_sky_map_
            self.mask = obj.mask
            self.convolver_image = obj.convolver_image
            self.grids = obj.grids
            self.border = obj.border


class FittingHyperImage(FittingImage):

    def __new__(cls, image, mask, hyper_model_image, hyper_galaxy_images, hyper_minimum_values, sub_grid_size=2,
                image_psf_shape=None):
        return np.array(mask.map_2d_array_to_masked_1d_array(image)).view(cls)

    def __init__(self, image, mask, hyper_model_image, hyper_galaxy_images, hyper_minimum_values, sub_grid_size=2,
                 image_psf_shape=None):
        """A fitting hyper regular is a fitting_image (see *FittingImage) which additionally includes a set of \
        'hyper_data'. This hyper-data is the best-fit model images of the observed data from a previous analysis, \
        and it is used to scale the noise in the regular, so as to avoid over-fitting localized regions of the regular \
        where the model does not provide a good fit.

        Look at the *FittingImage* docstring for all parameters / attributes not specific to a hyper regular.

        Parameters
        ----------
        hyper_model_images : [ndarray]
            List of the masked 1D array best-fit model regular's to each observed regular in a previous analysis.
        hyper_galaxy_images : [[ndarray]]
            List of the masked 1D array best-fit model regular's of every galaxy to each observed regular in a previous \
            analysis.

        Attributes
        ----------

        """
        super(FittingHyperImage, self).__init__(image=image, mask=mask, sub_grid_size=sub_grid_size,
                                                image_psf_shape=image_psf_shape)

        self.hyper_model_image = hyper_model_image
        self.hyper_galaxy_images = hyper_galaxy_images
        self.hyper_minimum_values = hyper_minimum_values

    def __array_finalize__(self, obj):
        super(FittingImage, self).__array_finalize__(obj)
        if isinstance(obj, FittingHyperImage):
            self.noise_map_ = obj.noise_map_
            self.background_noise_map_ = obj.background_noise_map_
            self.poisson_noise_map_ = obj.poisson_noise_map_
            self.exposure_time_map_ = obj.exposure_time_map_
            self.background_sky_map_ = obj.background_sky_map_
            self.mask = obj.mask
            self.convolver_image = obj.convolver_image
            self.grids = obj.grids
            self.border = obj.border
            self.hyper_model_image = obj.hyper_model_image
            self.hyper_galaxy_images = obj.hyper_galaxy_images
            self.hyper_minimum_values = obj.hyper_minimum_values


from autolens.data.array.plotters import array_plotters

def plot_intensities(light_profile, grid, mask=None, positions=None, as_subplot=False,
                     units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
                     cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                     cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                     title='Intensities', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                     mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                     output_path=None, output_format='show', output_filename='intensities'):

    intensities = light_profile.intensities_from_grid(grid=grid)
    intensities = grid.scaled_array_from_array_1d(intensities)

    array_plotters.plot_array(array=intensities, mask=mask, positions=positions, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                              grid_pointsize=grid_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_surface_density(mass_profile, grid, mask=None, positions=None, as_subplot=False,
                         units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
                         cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                         cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                         title='Surface Density', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                         mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                         output_path=None, output_format='show', output_filename='surface_density'):

    surface_density = mass_profile.surface_density_from_grid(grid=grid)
    surface_density = grid.scaled_array_from_array_1d(surface_density)

    array_plotters.plot_array(array=surface_density, mask=mask, positions=positions, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                              grid_pointsize=grid_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_potential(mass_profile, grid, mask=None, positions=None, as_subplot=False,
                   units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
                   cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                   cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                   title='Potential', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                   mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                   output_path=None, output_format='show', output_filename='potential'):

    potential = mass_profile.potential_from_grid(grid=grid)
    potential = grid.scaled_array_from_array_1d(potential)

    array_plotters.plot_array(array=potential, mask=mask, positions=positions, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                              grid_pointsize=grid_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)


def plot_deflections_y(mass_profile, grid, mask=None, positions=None, as_subplot=False,
                       units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
                       cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                       cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                       title='Deflections (y)', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                       mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                       output_path=None, output_format='show', output_filename='deflections_y'):
    
    deflections = mass_profile.deflections_from_grid(grid)
    deflections_y = grid.scaled_array_from_array_1d(deflections[:,0])

    array_plotters.plot_array(array=deflections_y, mask=mask, positions=positions, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                              grid_pointsize=grid_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)


def plot_deflections_x(mass_profile, grid, mask=None, positions=None, as_subplot=False,
                       units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
                       cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                       cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                       title='Deflections (x)', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                       mask_pointsize=10, position_pointsize=10.0, grid_pointsize=1,
                       output_path=None, output_format='show', output_filename='deflections_x'):

    deflections = mass_profile.deflections_from_grid(grid)
    deflections_x = grid.scaled_array_from_array_1d(deflections[:, 1])

    array_plotters.plot_array(array=deflections_x, mask=mask, positions=positions, as_subplot=as_subplot,
                              units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                              cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                              linthresh=linthresh, linscale=linscale,
                              cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                              title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                              xyticksize=xyticksize,
                              mask_pointsize=mask_pointsize, position_pointsize=position_pointsize,
                              grid_pointsize=grid_pointsize,
                              output_path=output_path, output_format=output_format, output_filename=output_filename)


import os

from autofit import conf
from autofit.core import non_linear as nl
from autolens.model.galaxy import galaxy, galaxy_model as gm
from autolens.model.inversion import pixelizations as pix, regularization as reg
from autolens.pipeline import phase as ph
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from test.integration import tools

dirpath = os.path.dirname(os.path.realpath(__file__))
dirpath = os.path.dirname(dirpath)
output_path = '{}/../output/lens_and_source_inversion'.format(dirpath)

path = '{}/../../'.format(os.path.dirname(os.path.realpath(__file__)))
conf.instance = conf.Config(config_path=path+'config', output_path=path+'output')

def pipeline():

    pipeline_name = "lens_mass_x1_source_x1_adaptive"
    data_name = '/lens_mass_x1_source_x1_adaptive'

    tools.reset_paths(data_name, pipeline_name, output_path)

    lens_mass = mp.EllipticalIsothermal(centre=(0.0, 0.0), axis_ratio=0.8, phi=80.0, einstein_radius=1.6)
    source_light = lp.EllipticalSersic(centre=(-0.0, 0.0), axis_ratio=0.6, phi=90.0, intensity=1.0,
                                       effective_radius=0.5, sersic_index=1.0)

    lens_galaxy = galaxy.Galaxy(sie=lens_mass)
    source_galaxy = galaxy.Galaxy(sersic=source_light)

    tools.simulate_integration_image(data_name=data_name, pixel_scale=0.2, lens_galaxies=[lens_galaxy],
                                     source_galaxies=[source_galaxy], target_signal_to_noise=30.0)

    pipeline = make_lens_x1_source_x1_inversion_pipeline(pipeline_name=pipeline_name)
    image = tools.load_image(data_name=data_name, pixel_scale=0.2)

    results = pipeline.run(image=image)
    for result in results:
        print(result)


def make_lens_x1_source_x1_inversion_pipeline(pipeline_name):
    class SourcePix(ph.LensSourcePlanePhase):

        def pass_priors(self, previous_results):
            self.lens_galaxies[0].sie.centre.centre_0 = 0.0
            self.lens_galaxies[0].sie.centre.centre_1 = 0.0
            self.lens_galaxies[0].sie.einstein_radius = 1.6
            self.source_galaxies[0].pixelization.shape_0 = 20.0
            self.source_galaxies[0].pixelization.shape_1 = 20.0

    phase1 = SourcePix(lens_galaxies=[gm.GalaxyModel(sie=mp.EllipticalIsothermal)],
                       source_galaxies=[gm.GalaxyModel(pixelization=pix.AdaptiveMagnification,
                                                       regularization=reg.Constant)],
                       optimizer_class=nl.MultiNest, phase_name="{}/phase1".format(pipeline_name))

    phase1.optimizer.n_live_points = 60
    phase1.optimizer.sampling_efficiency = 0.8

    return pl.PipelineImaging(pipeline_name, phase1)


if __name__ == "__main__":
    pipeline()
